# -*- coding: utf-8 -*-

import pytest

from py_gql._string_utils import dedent
from py_gql._utils import flatten
from py_gql.exc import SDLError
from py_gql.lang import ast as _ast
from py_gql.schema import UUID, Field, ObjectType, String
from py_gql.schema.build import SchemaDirective, make_executable_schema


def test_object_type_extension():
    assert (
        make_executable_schema(
            """
            type Query {
                foo: Object
            }

            interface IFace1 {
                one: String
            }

            type Object implements IFace1 {
                one: String
                two: Int
            }

            interface IFace2 {
                three: String
            }

            extend type Object implements IFace2 {
                three: String
            }
            """
        ).to_string()
        == dedent(
            """
            interface IFace1 {
                one: String
            }

            interface IFace2 {
                three: String
            }

            type Object implements IFace1 & IFace2 {
                one: String
                two: Int
                three: String
            }

            type Query {
                foo: Object
            }
            """
        )
    )


def test_injected_object_type_extension():
    Foo = ObjectType("Foo", [Field("one", String)])
    schema = make_executable_schema(
        """
        type Query {
            foo: Foo
        }

        extend type Foo {
            two: Int
        }
        """,
        additional_types=[Foo],
    )
    assert schema.to_string() == dedent(
        """
        type Foo {
            one: String
            two: Int
        }

        type Query {
            foo: Foo
        }
        """
    )

    assert schema.types["Foo"] is not Foo


def test_object_type_extension_duplicate_field():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo: String }
            type Object { one: String }
            extend type Object { one: Int }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 34, "line": 4}],
        "message": 'Found duplicate field "one" when extending type "Object"',
    }


def test_object_type_extension_already_implemented_interface():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo: String }
            interface IFace1 { one: String }
            type Object implements IFace1 { one: String }
            extend type Object implements IFace1
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 43, "line": 5}],
        "message": 'Interface "IFace1" already implemented for type "Object"',
    }


def test_object_type_extension_bad_extension():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo: String }
            type Object { one: String }
            extend input Object { two: Int }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 4}],
        "message": (
            "Expected ObjectTypeExtension when extending ObjectType "
            "but got InputObjectTypeExtension"
        ),
    }


def test_interface_type_extension():
    assert (
        make_executable_schema(
            """
            type Query {
                foo: IFace
            }

            interface IFace {
                one: String
            }

            extend interface IFace {
                two: String
            }
            """
        ).to_string()
        == dedent(
            """
            interface IFace {
                one: String
                two: String
            }

            type Query {
                foo: IFace
            }
            """
        )
    )


def test_interface_type_extension_duplicate_field():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo: IFace }
            interface IFace { one: String }
            extend interface IFace { one: Int }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 38, "line": 4}],
        "message": 'Found duplicate field "one" when extending interface "IFace"',
    }


def test_interface_type_extension_bad_extension():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo: IFace }
            interface IFace { one: String }
            extend type IFace { one: Int }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 4}],
        "message": (
            "Expected InterfaceTypeExtension when extending InterfaceType "
            "but got ObjectTypeExtension"
        ),
    }


def test_enum_extension():
    assert (
        make_executable_schema(
            """
            type Query {
                foo: Foo
            }

            enum Foo {
                BLUE
                GREEN
                RED
            }

            extend enum Foo {
                YELLOW
            }
            """
        ).to_string()
        == dedent(
            """
            enum Foo {
                BLUE
                GREEN
                RED
                YELLOW
            }

            type Query {
                foo: Foo
            }
            """
        )
    )


def test_enum_extension_duplicate_value():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query {
                foo: Foo
            }

            enum Foo {
                BLUE
                GREEN
                RED
            }

            extend enum Foo {
                RED
            }
            """
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 17, "line": 13}],
        "message": 'Found duplicate enum value "RED" when extending EnumType "Foo"',
    }


def test_enum_extension_bad_extension():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query {
                foo: Foo
            }

            enum Foo {
                BLUE
                GREEN
                RED
            }

            extend type Foo {
                one: Int
            }
            """
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 12}],
        "message": "Expected EnumTypeExtension when extending EnumType but got "
        "ObjectTypeExtension",
    }


def test_input_object_type_extension():
    assert (
        make_executable_schema(
            """
            type Query {
                foo(in: Foo): String
            }

            input Foo {
                one: Int
            }

            extend input Foo {
                two: String
            }
            """
        ).to_string()
        == dedent(
            """
            input Foo {
                one: Int
                two: String
            }

            type Query {
                foo(in: Foo): String
            }
            """
        )
    )


def test_input_object_type_extension_duplicate_field():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo(in: Foo): String }
            input Foo { one: Int }
            extend input Foo { one: Int }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 32, "line": 4}],
        "message": 'Found duplicate field "one" when extending input object "Foo"',
    }


def test_input_object_type_extension_bad_extension():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query { foo(in: Foo): String }
            input Foo { one: Int }
            extend type Foo { two: Int }
            """
        )
    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 4}],
        "message": (
            "Expected InputObjectTypeExtension when extending InputObjectType "
            "but got ObjectTypeExtension"
        ),
    }


def test_union_type_extension():
    assert (
        make_executable_schema(
            """
            type Query {
                foo: Foo
            }

            type Bar {
                bar: Int
            }

            type Baz {
                baz: Int
            }

            union Foo = Bar

            extend union Foo = Baz
            """
        ).to_string()
        == dedent(
            """
            type Bar {
                bar: Int
            }

            type Baz {
                baz: Int
            }

            union Foo = Bar | Baz

            type Query {
                foo: Foo
            }
            """
        )
    )


def test_union_type_extension_duplicate_type():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query {
                foo: Foo
            }

            type Bar {
                bar: Int
            }

            type Baz {
                baz: Int
            }

            union Foo = Bar

            extend union Foo = Bar
            """
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 32, "line": 16}],
        "message": (
            'Found duplicate member type "Bar" when extending UnionType "Foo"'
        ),
    }


def test_union_type_extension_bad_extension():
    with pytest.raises(SDLError) as exc_info:
        make_executable_schema(
            """
            type Query {
                foo: Foo
            }

            type Bar {
                bar: Int
            }

            type Baz {
                baz: Int
            }

            union Foo = Bar

            extend type Foo {
                one: Int
            }
            """
        )

    assert exc_info.value.to_dict() == {
        "locations": [{"column": 13, "line": 16}],
        "message": (
            "Expected UnionTypeExtension when extending UnionType "
            "but got ObjectTypeExtension"
        ),
    }


def test_scalar_type_extension():
    class ProtectedDirective(SchemaDirective):
        def visit_scalar(self, scalar_type):
            return scalar_type

    schema = make_executable_schema(
        """
        directive @protected on SCALAR

        type Query {
            foo: Foo
        }

        scalar Foo

        extend scalar Foo @protected
        """,
        schema_directives={"protected": ProtectedDirective},
    )

    assert schema.to_string() == dedent(
        """
        directive @protected on SCALAR

        scalar Foo

        type Query {
            foo: Foo
        }
        """
    )

    assert list(flatten(n.directives for n in schema.types["Foo"].nodes)) == [
        _ast.Directive(
            loc=(140, 150),
            name=_ast.Name(loc=(141, 150), value="protected"),
            arguments=[],
        )
    ]


def test_injected_scalar_type_extension():
    class ProtectedDirective(SchemaDirective):
        def visit_scalar(self, scalar_type):
            return scalar_type

    schema = make_executable_schema(
        """
        directive @protected on SCALAR

        type Query {
            foo: UUID
        }

        extend scalar UUID @protected
        """,
        additional_types=[UUID],
        schema_directives={"protected": ProtectedDirective},
    )

    assert schema.to_string(include_descriptions=False) == dedent(
        """
        directive @protected on SCALAR

        type Query {
            foo: UUID
        }

        scalar UUID
        """
    )

    assert list(flatten(n.directives for n in schema.types["UUID"].nodes)) == [
        _ast.Directive(
            loc=(122, 132),
            name=_ast.Name(loc=(123, 132), value="protected"),
            arguments=[],
        )
    ]

    assert schema.types["UUID"] is not UUID


def test_does_not_extend_specified_scalar():
    schema = make_executable_schema(
        """
        directive @protected on SCALAR

        type Query {
            foo: String
        }

        extend scalar String @protected
        """
    )

    assert schema.get_type("String") is String


def test_schema_extension():
    schema = make_executable_schema(
        """
        type Query { a: Boolean }

        type Foo { foo: String }
        type Bar { bar: String }

        extend schema {
            mutation: Bar
        }
        """
    )

    assert schema.query_type.name == "Query"
    assert schema.mutation_type.name == "Bar"
    assert schema.subscription_type is None


def test_schema_extension_directive():
    make_executable_schema(
        """
        directive @onSchema on SCHEMA

        type Foo { foo: String }
        type Bar { bar: String }

        schema {
            query: Foo
        }

        extend schema @onSchema
        """
    )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import NoUnusedFragmentsChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_all_fragment_names_are_used(schema):
    run_test(
        NoUnusedFragmentsChecker,
        schema,
        """
    {
        human(id: 4) {
            ...HumanFields1
            ... on Human {
                ...HumanFields2
            }
        }
    }
    fragment HumanFields1 on Human {
        name
        ...HumanFields3
    }
    fragment HumanFields2 on Human {
        name
    }
    fragment HumanFields3 on Human {
        name
    }
    """,
        [],
    )


def test_all_fragment_names_are_used_by_multiple_operations(schema):
    run_test(
        NoUnusedFragmentsChecker,
        schema,
        """
    query Foo {
        human(id: 4) {
            ...HumanFields1
        }
    }
    query Bar {
        human(id: 4) {
            ...HumanFields2
        }
    }
    fragment HumanFields1 on Human {
        name
        ...HumanFields3
    }
    fragment HumanFields2 on Human {
        name
    }
    fragment HumanFields3 on Human {
        name
    }
    """,
        [],
    )


def test_contains_unknown_fragments(schema):
    run_test(
        NoUnusedFragmentsChecker,
        schema,
        """
    query Foo {
        human(id: 4) {
            ...HumanFields1
        }
    }
    query Bar {
        human(id: 4) {
            ...HumanFields2
        }
    }
    fragment HumanFields1 on Human {
        name
        ...HumanFields3
    }
    fragment HumanFields2 on Human {
        name
    }
    fragment HumanFields3 on Human {
        name
    }
    fragment Unused1 on Human {
        name
    }
    fragment Unused2 on Human {
        name
    }
    """,
        ['Unused fragment(s) "Unused1", "Unused2"'],
    )


# Different from the reference test as we leave the cycle checks to
# the corresponding rule (NoFragmentCyclesChecker) while this rule
# considers used cyclic fragments as correct
def test_contains_unknown_fragments_with_ref_cycle(schema):
    run_test(
        NoUnusedFragmentsChecker,
        schema,
        """
    query Foo {
        human(id: 4) {
            ...HumanFields1
        }
    }
    query Bar {
        human(id: 4) {
            ...HumanFields2
        }
    }
    fragment HumanFields1 on Human {
        name
        ...HumanFields3
    }
    fragment HumanFields2 on Human {
        name
    }
    fragment HumanFields3 on Human {
        name
    }
    fragment Unused1 on Human {
        name
        ...Unused2
    }
    fragment Unused2 on Human {
        name
        ...Unused1
    }
    """,
    )


def test_contains_unknown_and_undef_fragments(schema):
    run_test(
        NoUnusedFragmentsChecker,
        schema,
        """
    query Foo {
        human(id: 4) {
            ...bar
        }
    }
    fragment foo on Human {
        name
    }
    """,
        ['Unused fragment(s) "foo"'],
    )



# -*- coding: utf-8 -*-


from py_gql.schema.scalars import Float, Int, String
from py_gql.schema.schema import Schema
from py_gql.schema.types import (
    Field,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    UnionType,
)


def test_references_are_equal():
    assert String == String


def test_int_and_float_are_not_equal():
    assert not Int == Float


def test_lists_of_same_type_are_equal():
    assert ListType(Int) == ListType(Int)


def test_lists_is_not_equal_to_item():
    assert not ListType(Int) == Int


def test_non_null_of_same_type_are_equal():
    assert NonNullType(Int) == NonNullType(Int)


def test_non_null_is_not_equal_to_nullable():
    assert not NonNullType(Int) == Int


def test_non_null_is_not_equal_to_list():
    assert not NonNullType(Int) == ListType(Int)


_schema = Schema(ObjectType("Query", [Field("field", String)]))


def test_same_referaence_is_subtype():
    assert _schema.is_subtype(String, String)


def test_int_is_not_subtype_of_float():
    assert not _schema.is_subtype(Int, Float)


def test_non_null_is_subtype_of_nullable():
    assert _schema.is_subtype(NonNullType(Int), Int)


def test_nullable_is_not_subtype_of_non_null():
    assert not _schema.is_subtype(Int, NonNullType(Int))


def test_item_is_not_subtype_of_list():
    assert not _schema.is_subtype(Int, ListType(Int))


def test_list_is_not_subtype_of_item():
    assert not _schema.is_subtype(ListType(Int), Int)


def test_list_of_non_null_is_subtype_of_list_of_nullable():
    assert _schema.is_subtype(ListType(NonNullType(Int)), ListType(Int))


def test_member_is_subtype_of_union():
    member = ObjectType("Object", [Field("field", String)])
    union = UnionType("Union", [member])
    schema = Schema(ObjectType("Query", [Field("field", union)]))
    assert schema.is_subtype(member, union)


def test_not_member_is_not_subtype_of_union():
    member = ObjectType("Object", [Field("field", String)])
    not_member = ObjectType("Object2", [Field("field", String)])
    union = UnionType("Union", [member])
    schema = Schema(ObjectType("Query", [Field("field", union)]))
    assert not schema.is_subtype(not_member, union)


def test_implementation_is_subtype_of_interface():
    iface = InterfaceType("Iface", [Field("field", String)])
    impl = ObjectType("Object", [Field("field", String)], [iface])
    schema = Schema(ObjectType("Query", [Field("field", impl)]))
    assert schema.is_subtype(impl, iface)


def test_not_implementation_is_not_subtype_of_interface():
    iface = InterfaceType("Iface", [Field("field", String)])
    impl = ObjectType("Object", [Field("field", String)])
    schema = Schema(ObjectType("Query", [Field("field", impl)]))
    assert not schema.is_subtype(impl, iface)


def test_references_overlap():
    assert _schema.types_overlap(Int, Int)


def test_int_and_float_do_not_overlap():
    assert not _schema.types_overlap(Int, Float)


def test_disjoint_unions_do_not_overlap():
    member1 = ObjectType("Object1", [Field("field", String)])
    member2 = ObjectType("Object2", [Field("field", String)])
    union1 = UnionType("Union1", [member2])
    union2 = UnionType("Union2", [member1])
    schema = Schema(
        ObjectType("Query", [Field("field1", union1), Field("field2", union2)])
    )
    assert not schema.types_overlap(union1, union2)


def test_common_unions_not_overlap():
    member1 = ObjectType("Object1", [Field("field", String)])
    member2 = ObjectType("Object2", [Field("field", String)])
    union1 = UnionType("Union1", [member2])
    union2 = UnionType("Union2", [member1, member2])
    schema = Schema(
        ObjectType("Query", [Field("field1", union1), Field("field2", union2)])
    )
    assert schema.types_overlap(union1, union2)


def test_union_and_interface_with_common_types_overlap():
    iface = InterfaceType("Iface", [Field("field", String)])
    impl = ObjectType("Object", [Field("field", String)], [iface])
    union = UnionType("Union", [impl])
    schema = Schema(
        ObjectType("Query", [Field("field1", union), Field("field2", iface)])
    )
    assert schema.types_overlap(iface, union)



# -*- coding: utf-8 -*-
""" Visitors are the basic abstraction used to traverse a GraphQL AST.

Note:
    While the concept is similar to the one used in the reference implementation,
    this doesn't support editing the ast, tracking of the visitor path or
    providing shared context. You can refer to :mod:`py_gql.validation` for
    example implementations.
"""

import functools as ft

from . import ast as _ast
from ..exc import GraphQLError

__all__ = (
    "SkipNode",
    "Visitor",
    "visit",
    "DispatchingVisitor",
    "ParrallelVisitor",
)


class SkipNode(GraphQLError):
    def __init__(self):
        super(SkipNode, self).__init__("")


class Visitor(object):
    """ Visitor base class. """

    def enter(self, node):
        """ Function called when entering a node.

        Raising :class:`SkipNode` in this function will lead to ignoring the
        current node and all it's descendants. as well as prevent running
        :meth:`leave`.

        Args:
            node (py_gql.lang.ast.Node) Node being visited
        """
        pass

    def leave(self, node):
        """ Function called when leaving a node. Will not be called if
        :class:`SkipNode` was raised in `enter`.

        Args:
            node (py_gql.lang.ast.Node) Node being visited
        """
        pass


def visit(visitor, ast_root):
    """ Visit a GraphQL parse tree.

    All side effects and results of traversal should be contained in the
    visitor instance.

    Args:
        visitor (Visitor): Visitor instance
        ast_root (py_gql.lang.ast.Document): Parsed GraphQL document
    """
    assert isinstance(ast_root, _ast.Document)
    _visit_document(visitor, ast_root)


def _visiting(func):
    """ Wrap a function to call the provided visitor ``enter`` / ``leave``
    functions before / after processing a node.
    """

    @ft.wraps(func)
    def wrapper(visitor, node):
        if node is None:
            return

        try:
            visitor.enter(node)
        except SkipNode:
            return

        func(visitor, node)
        visitor.leave(node)

    return wrapper


def _many(fn, visitor, nodes):
    """ Run a visiting procedure on multiple nodes """
    if nodes:
        for node in nodes:
            fn(visitor, node)


@_visiting
def _visit_document(visitor, document):
    for definition in document.definitions:
        _visit_definition(visitor, definition)


def _visit_definition(visitor, definition):
    func = {
        _ast.OperationDefinition: _visit_operation_definition,
        _ast.FragmentDefinition: _visit_fragment_definition,
        _ast.SchemaDefinition: _visit_schema_definition,
        _ast.ScalarTypeDefinition: _visit_scalar_type_definition,
        _ast.ObjectTypeDefinition: _visit_object_type_definition,
        _ast.InterfaceTypeDefinition: _visit_interface_type_definition,
        _ast.UnionTypeDefinition: _visit_union_type_definition,
        _ast.EnumTypeDefinition: visit_enum_type_definition,
        _ast.InputObjectTypeDefinition: _visit_input_object_type_definition,
        _ast.SchemaExtension: _visit_schema_definition,
        _ast.ScalarTypeExtension: _visit_scalar_type_definition,
        _ast.ObjectTypeExtension: _visit_object_type_definition,
        _ast.InterfaceTypeExtension: _visit_interface_type_definition,
        _ast.UnionTypeExtension: _visit_union_type_definition,
        _ast.EnumTypeExtension: visit_enum_type_definition,
        _ast.InputObjectTypeExtension: _visit_input_object_type_definition,
        _ast.DirectiveDefinition: _visit_directive_definition,
    }.get(type(definition), None)
    if func is not None:
        func(visitor, definition)


@_visiting
def _visit_operation_definition(visitor, definition):
    _many(_visit_variable_definition, visitor, definition.variable_definitions)
    _many(_visit_directive, visitor, definition.directives)
    _visit_selection_set(visitor, definition.selection_set)


@_visiting
def _visit_fragment_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)
    _visit_selection_set(visitor, definition.selection_set)


@_visiting
def _visit_variable_definition(visitor, variable_definition):
    if variable_definition.default_value:
        _visit_input_value(visitor, variable_definition.default_value)
    _visit_named_type(visitor, variable_definition.type)


@_visiting
def _visit_named_type(visitor, named_type):
    pass


@_visiting
def _visit_directive(visitor, directive):
    _many(_visit_argument, visitor, directive.arguments)


@_visiting
def _visit_argument(visitor, argument):
    _visit_input_value(visitor, argument.value)


@_visiting
def _visit_selection_set(visitor, selection_set):
    _many(_visit_selection, visitor, selection_set.selections)


def _visit_selection(visitor, selection):
    func = {
        _ast.Field: _visit_field,
        _ast.FragmentSpread: _visit_fragment_spread,
        _ast.InlineFragment: _visit_inline_fragment,
    }[type(selection)]
    func(visitor, selection)


@_visiting
def _visit_field(visitor, field):
    _many(_visit_argument, visitor, field.arguments)
    _many(_visit_directive, visitor, field.directives)
    _visit_selection_set(visitor, field.selection_set)


@_visiting
def _visit_fragment_spread(visitor, spread):
    _many(_visit_directive, visitor, spread.directives)


@_visiting
def _visit_inline_fragment(visitor, fragment):
    _many(_visit_directive, visitor, fragment.directives)
    if fragment.selection_set:
        _visit_selection_set(visitor, fragment.selection_set)


@_visiting
def _visit_input_value(visitor, input_value):
    kind = type(input_value)
    if kind == _ast.ObjectValue:
        _many(_visit_object_field, visitor, input_value.fields)
    elif kind == _ast.ListValue:
        _many(_visit_input_value, visitor, input_value.values)


@_visiting
def _visit_object_field(visitor, field):
    _visit_input_value(visitor, field.value)


@_visiting
def _visit_schema_definition(visitor, definition):
    _many(_visit_operation_type_definition, visitor, definition.operation_types)
    _many(_visit_directive, visitor, definition.directives)


@_visiting
def _visit_operation_type_definition(visitor, definition):
    _visit_named_type(visitor, definition.type)


@_visiting
def _visit_scalar_type_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)


@_visiting
def _visit_object_type_definition(visitor, definition):
    _many(_visit_named_type, visitor, definition.interfaces)
    _many(_visit_directive, visitor, definition.directives)
    _many(_visit_field_definition, visitor, definition.fields)


@_visiting
def _visit_interface_type_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)
    _many(_visit_field_definition, visitor, definition.fields)


@_visiting
def _visit_union_type_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)
    _many(_visit_named_type, visitor, definition.types)


@_visiting
def visit_enum_type_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)
    _many(_visit_enum_value_definition, visitor, definition.values)


@_visiting
def _visit_input_object_type_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)
    _many(_visit_input_value_definition, visitor, definition.fields)


@_visiting
def _visit_field_definition(visitor, definition):
    _visit_named_type(visitor, definition.type)
    _many(_visit_input_value_definition, visitor, definition.arguments)
    _many(_visit_directive, visitor, definition.directives)


@_visiting
def _visit_input_value_definition(visitor, definition):
    _visit_named_type(visitor, definition.type)
    _visit_input_value(visitor, definition.default_value)
    _many(_visit_directive, visitor, definition.directives)


@_visiting
def _visit_enum_value_definition(visitor, definition):
    _many(_visit_directive, visitor, definition.directives)


@_visiting
def _visit_directive_definition(visitor, definition):
    _many(_visit_input_value_definition, visitor, definition.arguments)


class DispatchingVisitor(Visitor):
    """ Base class for specialised visitors.

    You should subclass this and implement methods named ``enter_*`` and
    ``leave_*`` where ``*`` represents the node class to be handled.
    For instance to process :class:`py_gql.lang.ast.FloatValue` nodes,
    implement ``enter_float_value``.

    Default behaviour is noop for all classes.
    """

    def enter(self, node):  # noqa: C901
        kind = type(node)
        if kind is _ast.Document:
            self.enter_document(node)
        elif kind is _ast.OperationDefinition:
            self.enter_operation_definition(node)
        elif kind is _ast.FragmentDefinition:
            self.enter_fragment_definition(node)
        elif kind is _ast.VariableDefinition:
            self.enter_variable_definition(node)
        elif kind is _ast.Directive:
            self.enter_directive(node)
        elif kind is _ast.Argument:
            self.enter_argument(node)
        elif kind is _ast.SelectionSet:
            self.enter_selection_set(node)
        elif kind is _ast.Field:
            self.enter_field(node)
        elif kind is _ast.FragmentSpread:
            self.enter_fragment_spread(node)
        elif kind is _ast.InlineFragment:
            self.enter_inline_fragment(node)
        elif kind is _ast.NullValue:
            self.enter_null_value(node)
        elif kind is _ast.IntValue:
            self.enter_int_value(node)
        elif kind is _ast.FloatValue:
            self.enter_float_value(node)
        elif kind is _ast.StringValue:
            self.enter_string_value(node)
        elif kind is _ast.BooleanValue:
            self.enter_boolean_value(node)
        elif kind is _ast.EnumValue:
            self.enter_enum_value(node)
        elif kind is _ast.Variable:
            self.enter_variable(node)
        elif kind is _ast.ListValue:
            self.enter_list_value(node)
        elif kind is _ast.ObjectValue:
            self.enter_object_value(node)
        elif kind is _ast.ObjectField:
            self.enter_object_field(node)
        elif kind is _ast.NamedType:
            self.enter_named_type(node)
        elif kind is _ast.ListType:
            self.enter_list_type(node)
        elif kind is _ast.NonNullType:
            self.enter_non_null_type(node)
        elif kind is _ast.SchemaDefinition:
            self.enter_schema_definition(node)
        elif kind is _ast.OperationTypeDefinition:
            self.enter_operation_type_definition(node)
        elif kind is _ast.ScalarTypeDefinition:
            self.enter_scalar_type_definition(node)
        elif kind is _ast.ObjectTypeDefinition:
            self.enter_object_type_definition(node)
        elif kind is _ast.FieldDefinition:
            self.enter_field_definition(node)
        elif kind is _ast.InputValueDefinition:
            self.enter_input_value_definition(node)
        elif kind is _ast.InterfaceTypeDefinition:
            self.enter_interface_type_definition(node)
        elif kind is _ast.UnionTypeDefinition:
            self.enter_union_type_definition(node)
        elif kind is _ast.EnumTypeDefinition:
            self.enter_enum_type_definition(node)
        elif kind is _ast.EnumValueDefinition:
            self.enter_enum_value_definition(node)
        elif kind is _ast.InputObjectTypeDefinition:
            self.enter_input_object_type_definition(node)
        elif kind is _ast.SchemaExtension:
            self.enter_schema_extension(node)
        elif kind is _ast.ScalarTypeExtension:
            self.enter_scalar_type_extension(node)
        elif kind is _ast.ObjectTypeExtension:
            self.enter_object_type_extension(node)
        elif kind is _ast.InterfaceTypeExtension:
            self.enter_interface_type_extension(node)
        elif kind is _ast.UnionTypeExtension:
            self.enter_union_type_extension(node)
        elif kind is _ast.EnumTypeExtension:
            self.enter_enum_type_extension(node)
        elif kind is _ast.InputObjectTypeExtension:
            self.enter_input_object_type_extension(node)
        elif kind is _ast.DirectiveDefinition:
            self.enter_directive_definition(node)

    def leave(self, node):  # noqa: C901
        kind = type(node)
        if kind is _ast.Document:
            self.leave_document(node)
        elif kind is _ast.OperationDefinition:
            self.leave_operation_definition(node)
        elif kind is _ast.FragmentDefinition:
            self.leave_fragment_definition(node)
        elif kind is _ast.VariableDefinition:
            self.leave_variable_definition(node)
        elif kind is _ast.Directive:
            self.leave_directive(node)
        elif kind is _ast.Argument:
            self.leave_argument(node)
        elif kind is _ast.SelectionSet:
            self.leave_selection_set(node)
        elif kind is _ast.Field:
            self.leave_field(node)
        elif kind is _ast.FragmentSpread:
            self.leave_fragment_spread(node)
        elif kind is _ast.InlineFragment:
            self.leave_inline_fragment(node)
        elif kind is _ast.NullValue:
            self.leave_null_value(node)
        elif kind is _ast.IntValue:
            self.leave_int_value(node)
        elif kind is _ast.FloatValue:
            self.leave_float_value(node)
        elif kind is _ast.StringValue:
            self.leave_string_value(node)
        elif kind is _ast.BooleanValue:
            self.leave_boolean_value(node)
        elif kind is _ast.EnumValue:
            self.leave_enum_value(node)
        elif kind is _ast.Variable:
            self.leave_variable(node)
        elif kind is _ast.ListValue:
            self.leave_list_value(node)
        elif kind is _ast.ObjectValue:
            self.leave_object_value(node)
        elif kind is _ast.ObjectField:
            self.leave_object_field(node)
        elif kind is _ast.NamedType:
            self.leave_named_type(node)
        elif kind is _ast.ListType:
            self.leave_list_type(node)
        elif kind is _ast.NonNullType:
            self.leave_non_null_type(node)
        elif kind is _ast.SchemaDefinition:
            self.leave_schema_definition(node)
        elif kind is _ast.OperationTypeDefinition:
            self.leave_operation_type_definition(node)
        elif kind is _ast.ScalarTypeDefinition:
            self.leave_scalar_type_definition(node)
        elif kind is _ast.ObjectTypeDefinition:
            self.leave_object_type_definition(node)
        elif kind is _ast.FieldDefinition:
            self.leave_field_definition(node)
        elif kind is _ast.InputValueDefinition:
            self.leave_input_value_definition(node)
        elif kind is _ast.InterfaceTypeDefinition:
            self.leave_interface_type_definition(node)
        elif kind is _ast.UnionTypeDefinition:
            self.leave_union_type_definition(node)
        elif kind is _ast.EnumTypeDefinition:
            self.leave_enum_type_definition(node)
        elif kind is _ast.EnumValueDefinition:
            self.leave_enum_value_definition(node)
        elif kind is _ast.InputObjectTypeDefinition:
            self.leave_input_object_type_definition(node)
        elif kind is _ast.SchemaExtension:
            self.leave_schema_extension(node)
        elif kind is _ast.ScalarTypeExtension:
            self.leave_scalar_type_extension(node)
        elif kind is _ast.ObjectTypeExtension:
            self.leave_object_type_extension(node)
        elif kind is _ast.InterfaceTypeExtension:
            self.leave_interface_type_extension(node)
        elif kind is _ast.UnionTypeExtension:
            self.leave_union_type_extension(node)
        elif kind is _ast.EnumTypeExtension:
            self.leave_enum_type_extension(node)
        elif kind is _ast.InputObjectTypeExtension:
            self.leave_input_object_type_extension(node)
        elif kind is _ast.DirectiveDefinition:
            self.leave_directive_definition(node)

    def _default_handler(self, node):
        pass

    enter_document = _default_handler
    leave_document = _default_handler

    enter_operation_definition = _default_handler
    leave_operation_definition = _default_handler

    enter_fragment_definition = _default_handler
    leave_fragment_definition = _default_handler

    enter_variable_definition = _default_handler
    leave_variable_definition = _default_handler

    enter_directive = _default_handler
    leave_directive = _default_handler

    enter_argument = _default_handler
    leave_argument = _default_handler

    enter_selection_set = _default_handler
    leave_selection_set = _default_handler

    enter_field = _default_handler
    leave_field = _default_handler

    enter_fragment_spread = _default_handler
    leave_fragment_spread = _default_handler

    enter_inline_fragment = _default_handler
    leave_inline_fragment = _default_handler

    enter_null_value = _default_handler
    leave_null_value = _default_handler

    enter_int_value = _default_handler
    leave_int_value = _default_handler

    enter_float_value = _default_handler
    leave_float_value = _default_handler

    enter_string_value = _default_handler
    leave_string_value = _default_handler

    enter_boolean_value = _default_handler
    leave_boolean_value = _default_handler

    enter_enum_value = _default_handler
    leave_enum_value = _default_handler

    enter_variable = _default_handler
    leave_variable = _default_handler

    enter_list_value = _default_handler
    leave_list_value = _default_handler

    enter_object_value = _default_handler
    leave_object_value = _default_handler

    enter_object_field = _default_handler
    leave_object_field = _default_handler

    enter_named_type = _default_handler
    leave_named_type = _default_handler

    enter_list_type = _default_handler
    leave_list_type = _default_handler

    enter_non_null_type = _default_handler
    leave_non_null_type = _default_handler

    enter_schema_definition = _default_handler
    leave_schema_definition = _default_handler

    enter_operation_type_definition = _default_handler
    leave_operation_type_definition = _default_handler

    enter_scalar_type_definition = _default_handler
    leave_scalar_type_definition = _default_handler

    enter_object_type_definition = _default_handler
    leave_object_type_definition = _default_handler

    enter_field_definition = _default_handler
    leave_field_definition = _default_handler

    enter_input_value_definition = _default_handler
    leave_input_value_definition = _default_handler

    enter_interface_type_definition = _default_handler
    leave_interface_type_definition = _default_handler

    enter_union_type_definition = _default_handler
    leave_union_type_definition = _default_handler

    enter_enum_type_definition = _default_handler
    leave_enum_type_definition = _default_handler

    enter_enum_value_definition = _default_handler
    leave_enum_value_definition = _default_handler

    enter_input_object_type_definition = _default_handler
    leave_input_object_type_definition = _default_handler

    enter_schema_extension = _default_handler
    leave_schema_extension = _default_handler

    enter_scalar_type_extension = _default_handler
    leave_scalar_type_extension = _default_handler

    enter_object_type_extension = _default_handler
    leave_object_type_extension = _default_handler

    enter_interface_type_extension = _default_handler
    leave_interface_type_extension = _default_handler

    enter_union_type_extension = _default_handler
    leave_union_type_extension = _default_handler

    enter_enum_type_extension = _default_handler
    leave_enum_type_extension = _default_handler

    enter_input_object_type_extension = _default_handler
    leave_input_object_type_extension = _default_handler

    enter_directive_definition = _default_handler
    leave_directive_definition = _default_handler


class ParrallelVisitor(Visitor):
    """ Abstraction to run multiple visitor instances as one.

    - All visitors are run in the order they are defined
    - raising :class:`SkipNode` in one of them will prevent any later visitor
      to run

    Args:
        *visitors (List[Visitor]): List of visitors to run
    """

    # pylint: disable = super-init-not-called
    def __init__(self, *visitors):
        self.visitors = visitors

    def enter(self, node):
        for v in self.visitors:
            v.enter(node)

    def leave(self, node):
        for v in self.visitors[::-1]:
            v.leave(node)



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import VariablesAreInputTypesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_input_types_are_valid(schema):
    run_test(
        VariablesAreInputTypesChecker,
        schema,
        """
    query Foo($a: String, $b: [Boolean!]!, $c: ComplexInput) {
        field(a: $a, b: $b, c: $c)
    }
    """,
    )


def test_output_types_are_invalid(schema):
    run_test(
        VariablesAreInputTypesChecker,
        schema,
        """
    query Foo($a: Dog, $b: [[CatOrDog!]]!, $c: Pet) {
        field(a: $a, b: $b, c: $c)
    }
    """,
        [
            'Variable "$a" must be input type',
            'Variable "$b" must be input type',
            'Variable "$c" must be input type',
        ],
    )


def test_unknown_types_are_invalid(schema):
    run_test(
        VariablesAreInputTypesChecker,
        schema,
        """
    query Foo($a: Bar) {
        field(a: $a)
    }
    """,
        ['Variable "$a" must be input type'],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=all
"""
"""

import imp
import itertools
import os
import sys

from setuptools import find_packages, setup

ABBOUT = imp.load_source("about", os.path.join(".", "py_gql", "__version__.py"))


def run_setup():

    with open("README.md") as f:
        readme = "\n" + f.read()

    setup(
        name=ABBOUT.__title__,
        version=ABBOUT.__version__,
        description=ABBOUT.__description__,
        long_description=readme,
        long_description_content_type="text/markdown",
        author=ABBOUT.__author__,
        author_email=ABBOUT.__author_email__,
        url=ABBOUT.__url__,
        license=ABBOUT.__license__,
        keywords="graphql",
        zip_safe=False,
        packages=find_packages(
            exclude=("tests", "tests.*", "docs", "examples")
        ),
        # package_data={},
        install_requires=_split_requirements("requirements.txt"),
        include_package_data=True,
        python_requires=">2.6, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
        ext_modules=_cython_ext_modules(
            "py_gql",
            "py_gql.lang",
            "py_gql.validation",
            "py_gql.validation.rules",
            "py_gql.utilities",
            "py_gql.schema",
            "py_gql.schema.build",
            "py_gql.execution",
        ),
        classifiers=[
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Programming Language :: Python",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Operating System :: POSIX",
            "Operating System :: MacOS :: MacOS X",
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Topic :: Software Development :: Libraries",
        ],
        tests_require=(
            _split_requirements(
                "test-requirements.txt", "py3-test-requirements.txt"
            )
            if sys.version >= "3"
            else _split_requirements("test-requirements.txt")
        ),
    )


def _cython_ext_modules(*packages):
    try:
        from Cython.Build import cythonize
    except ImportError:
        return []
    else:
        enable_linetrace = "CYTHON_TRACE" in os.environ
        ext_modules = [
            cythonize(
                "%s/*.py" % package.replace(".", "/"),
                compiler_directives={
                    "embedsignature": True,
                    "linetrace": enable_linetrace,
                },
            )
            for package in packages
        ]
        return list(itertools.chain.from_iterable(ext_modules))


def _split_requirements(*requirements_files):
    req = []
    for requirements_file in requirements_files:
        with open(requirements_file) as f:
            req.extend(
                [
                    line.strip()
                    for line in f.readlines()
                    if line.strip() and not line.startswith("#")
                ]
            )
    return req


if __name__ == "__main__":
    run_setup()



# -*- coding: utf-8 -*-
""" Schema definition. """

from collections import defaultdict

from ..exc import SchemaError, UnknownType
from ..lang import ast as _ast
from ._validation import validate_schema
from .directives import SPECIFIED_DIRECTIVES
from .introspection import __Schema__
from .printer import SchemaPrinter
from .scalars import SPECIFIED_SCALAR_TYPES
from .types import (
    Directive,
    InputObjectType,
    InterfaceType,
    ListType,
    NonNullType,
    ObjectType,
    Type,
    UnionType,
    WrappingType,
    is_abstract_type,
    unwrap_type,
)

_unset = object()


class Schema(object):
    """ A GraphQL schema definition.

    This is the main container for a GraphQL schema and its related types.

    Args:
        query_type (py_gql.schema.ObjectType):
            The root query type for the schema

        mutation_type (py_gql.schema.ObjectType):
            The root mutation type for the schema

        subscription_type (py_gql.schema.ObjectType):
            The root subscription type for the schema

        directives (List[py_gql.schema.Directive]):
            List of possible directives to use.
            The default, specified directives (``@include``, ``@skip``) will
            always be included.

        types (List[py_gql.schema.Type]):
            List of additional types on top of the types that would be infered
            from the root types.

        node (List[Union[py_gql.lang.ast.SchemaDefinition, \
                         py_gql.lang.ast.SchemaExtension]]):
            AST node for the schema if applicable, i.e. when creating the schema
            from a GraphQL (SDL) document.

    Attributes:
        query_type (py_gql.schema.ObjectType):
            The root query type for the schema (required)

        mutation_type (py_gql.schema.ObjectType):
            The root mutation type for the schema (optional)

        subscription_type (py_gql.schema.ObjectType):
            The root subscription type for the schema (optional)

        node (List[Union[py_gql.lang.ast.SchemaDefinition, \
                         py_gql.lang.ast.SchemaExtension]]):
            AST node for the schema if applicable, i.e. when creating the schema
            from a GraphQL (SDL) document.

        type_map (Dict[str, py_gql.schema.Type]):
            Mapping ``type name -> Type instance`` of all types used in the
            schema, including directives.

        types (Dict[str, py_gql.schema.Type]):
            Mapping ``type name -> Type instance`` of all types used in the
            schema, excluding directives.

        directives (Dict[str, py_gql.schema.Directive]):
            Mapping ``directive name -> Directive instance`` of all directives
            used in the schema.

        implementations (Dict[str, List[py_gql.schema.Directive]]):
            Mapping of ``interface name -> [implementing object types]``.
    """

    def __init__(
        self,
        query_type=None,
        mutation_type=None,
        subscription_type=None,
        directives=None,
        types=None,
        nodes=None,
    ):
        self.query_type = query_type
        self.mutation_type = mutation_type
        self.subscription_type = subscription_type

        self._types = [query_type, mutation_type, subscription_type]
        self._types.append(__Schema__)
        self.nodes = nodes or []

        # NOTE: This is the notion of the specified types being always
        # available. As a result of this line, intropection queries will always
        # include thes types even if they are not actively used in the schema.
        # I am not totally sure if this is the right behaviour and it may
        # change in the future.
        self._types.extend(list(SPECIFIED_SCALAR_TYPES))

        if types:
            self._types.extend(types)

        self._directives = []
        if directives:
            self._directives.extend(directives)
        _directive_names = set(
            (d.name for d in self._directives if isinstance(d, Directive))
        )
        for d in SPECIFIED_DIRECTIVES:
            if d.name not in _directive_names:
                self._directives.append(d)

        self._possible_types = {}
        self._is_valid = None
        self._literal_types_cache = {}

        self.type_map = _build_type_map(self._types + self._directives)
        self._rebuild_caches()

    def _rebuild_caches(self):
        self.types = {
            name: t
            for name, t in self.type_map.items()
            if not isinstance(t, Directive)
        }
        self.directives = {
            name: t
            for name, t in self.type_map.items()
            if isinstance(t, Directive)
        }
        self.implementations = defaultdict(list)
        for type_ in self.types.values():
            if isinstance(type_, ObjectType):
                for iface in type_.interfaces:
                    self.implementations[iface.name].append(type_)
        self._is_valid = None

    def validate(self):
        """ Check that the schema is valid and cache the result.

        Returns:
            bool: ``True`` if the schema is valid

        Raises:
            :class:`~py_gql.exc.SchemaError` if the schema is invalid.
        """
        if self._is_valid is None:
            self._is_valid = validate_schema(self)
        return self._is_valid

    def get_type(self, name, default=_unset):
        """ Get a type by name.

        Args:
            name (str): Requested type name
            default: If set, will be returned in case there is no matching type

        Returns:
            py_gql.schema.Type: Type instance

        Raises:
            :class:`~py_gql.exc.UnknownType`:
                if ``default`` is not set and the type is not found

        """
        try:
            return self.types[name]
        except KeyError:
            if default is not _unset:
                return default
            raise UnknownType(name)

    def get_type_from_literal(self, ast_node):
        """ Given an AST node describing a type, return a
        corresponding :class:`py_gql.schema.Type` instance.

        For example, if provided the parsed AST node for ``[User]``,
        a :class:`py_gql.schema.ListType` instance will be returned, containing
        the type called ``User`` found in the schema.
        If a type called ``User`` is not found in the schema, then
        :class:`~py_gql.exc.UnknownType` will be raised.

        Args:
            ast_node (py_gql.lang.ast.Type)

        Returns:
            py_gql.schema.Type: Corresponding type instance

        Raises:
            :class:`~py_gql.exc.UnknownType`: if  any named type is not found
        """
        if ast_node in self._literal_types_cache:
            return self._literal_types_cache[ast_node]

        if isinstance(ast_node, _ast.ListType):
            t = ListType(self.get_type_from_literal(ast_node.type))
            self._literal_types_cache[ast_node] = t
            return t
        elif isinstance(ast_node, _ast.NonNullType):
            t = NonNullType(self.get_type_from_literal(ast_node.type))
            self._literal_types_cache[ast_node] = t
            return t
        elif isinstance(ast_node, _ast.NamedType):
            t = self.get_type(ast_node.name.value)
            self._literal_types_cache[ast_node] = t
            return t
        raise TypeError("Invalid type node %r" % ast_node)

    def get_possible_types(self, type_):
        """ Get the possible implementations of an abstract type.

        Args:
            type_ (Union[py_gql.schema.types.UnionType, \
py_gql.schema.types.InterfaceType]):
                Abstract type to check.

        Returns:
            List[py_gql.schema.types.ObjectType]: List of possible implementations.
        """
        if type_ in self._possible_types:
            return self._possible_types[type_]

        if isinstance(type_, UnionType):
            self._possible_types[type_] = type_.types or []
            return self._possible_types[type_]
        elif isinstance(type_, InterfaceType):
            self._possible_types[type_] = self.implementations.get(
                type_.name, []
            )
            return self._possible_types[type_]

        raise TypeError("Not an abstract type: %s" % type_)

    def is_possible_type(self, abstract_type, type_):
        """ Check that ``type_`` is a possible realization of ``abstract_type``.

        Args:
            abstract_type (Union[py_gql.schema.types.UnionType, \
py_gql.schema.types.InterfaceType]):
            type_ (py_gql.schema.Type): Concrete type to check

        Returns:
            bool: ``True`` if ``type_`` is valid for ``abstract_type``
        """
        return type_ in self.get_possible_types(abstract_type)

    def is_subtype(self, type_, super_type):
        """ Provided a type and a super type, return true if the first type is
        either equal or a subset of the second super type (covariant).

        Args:
            type_ (py_gql.schema.Type):
            super_type (py_gql.schema.Type):

        Returns:
            bool:

        """
        if type_ == super_type:
            return True

        if (
            isinstance(type_, WrappingType)
            and isinstance(super_type, WrappingType)
            and type(type_) == type(super_type)
        ):
            return self.is_subtype(type_.type, super_type.type)

        if isinstance(type_, NonNullType):
            return self.is_subtype(type_.type, super_type)

        if isinstance(type_, ListType):
            return False

        return (
            is_abstract_type(super_type)
            and isinstance(type_, ObjectType)
            and self.is_possible_type(super_type, type_)
        )

    def types_overlap(self, rhs, lhs):
        """ Provided two composite types, determine if they "overlap". Two
        composite types overlap when the Sets of possible concrete types for
        each intersect.

        This is often used to determine if a fragment of a given type could
        possibly be visited in a context of another type. This function is
        commutative.

        Args:
            rhs (py_gql.schema.Type):
            lhs (py_gql.schema.Type):

        Returns:
            bool:
        """
        if rhs == lhs:
            return True

        if is_abstract_type(rhs) and is_abstract_type(lhs):
            rhs_types = self.get_possible_types(rhs)
            lhs_types = self.get_possible_types(lhs)
            return any((t in lhs_types for t in rhs_types))

        return (is_abstract_type(rhs) and self.is_possible_type(rhs, lhs)) or (
            is_abstract_type(lhs) and self.is_possible_type(lhs, rhs)
        )

    def to_string(
        self,
        indent=4,
        include_descriptions=True,
        description_format="block",
        include_introspection=False,
        cls=SchemaPrinter,
    ):
        """ Format the schema as a string

        Args:
            indent (Union[str, int]): Indent character or number of spaces

            include_descriptions (bool):
                If ``True`` include descriptions in the output

            description_format ("comments"|"block"):
                Control how descriptions are formatted. ``"comments"`` is the
                old standard and will be compatible with most GraphQL parsers
                while ``"block"`` is part of the most recent specification and
                includes descriptions as block strings that can be extracted
                according to the specification.

            include_introspection (bool):
                If ``True``, include introspection types in the output

            cls (callable): Custom formatter
                Use this to customize the behaviour, by default this uses
                :class:`py_gql.schema.printer.SchemaPrinter`.

        Returns:
            str: Formatted GraphQL schema
        """
        return cls(
            indent=indent,
            include_descriptions=include_descriptions,
            description_format=description_format,
            include_introspection=include_introspection,
        )(self)


def _build_type_map(types, _type_map=None):
    """ Recursively build a mapping name <> Type from a list of types to include
    all referenced types.

    Warning:
        This will flatten all lazy type definitions and attributes.

    Args:
        types (List[py_gql.schema.Type]): List of types
        _type_map (Dict[str, py_gql.schema.Type]):
            Pre-built type map ( recursive calls)

    Returns:
        Dict[str, py_gql.schema.Type]
    """
    type_map = _type_map or {}
    for type_ in types or []:
        if not type_:
            continue

        type_ = unwrap_type(type_)

        if not (isinstance(type_, Type) and hasattr(type_, "name")):
            raise SchemaError(
                'Expected named types but got "%s" of type %s'
                % (type_, type(type_))
            )

        name = type_.name
        if name in type_map:
            if type_ is not type_map[name]:
                raise SchemaError('Duplicate type "%s"' % name)
            continue

        type_map[name] = type_
        child_types = []

        if isinstance(type_, UnionType):
            child_types.extend(type_.types)

        if isinstance(type_, ObjectType):
            child_types.extend(type_.interfaces)

        if isinstance(type_, (ObjectType, InterfaceType)):
            for field in type_.fields:
                child_types.append(field.type)
                child_types.extend([arg.type for arg in field.args or []])

        if isinstance(type_, InputObjectType):
            for field in type_.fields:
                child_types.append(field.type)

        if isinstance(type_, Directive):
            child_types.extend([arg.type for arg in type_.args or []])

        type_map.update(_build_type_map(child_types, _type_map=type_map))

    return type_map



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

from py_gql.validation.rules import OverlappingFieldsCanBeMergedChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_unique_fields(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment uniqueFields on Dog {
            name
            nickname
        }
        """,
    )


def test_identical_fields(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment mergeIdenticalFields on Dog {
            name
            name
        }
        """,
    )


def test_identical_fields_with_identical_args(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment mergeIdenticalFieldsWithIdenticalArgs on Dog {
            doesKnowCommand(dogCommand: SIT)
            doesKnowCommand(dogCommand: SIT)
        }
        """,
    )


def test_identical_fields_with_identical_directives(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment mergeSameFieldsWithSameDirectives on Dog {
            name @include(if: true)
            name @include(if: true)
        }
        """,
    )


def test_different_args_with_different_aliases(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment differentArgsWithDifferentAliases on Dog {
            knowsSit: doesKnowCommand(dogCommand: SIT)
            knowsDown: doesKnowCommand(dogCommand: DOWN)
        }
        """,
    )


def test_different_directives_with_different_aliases(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment differentDirectivesWithDifferentAliases on Dog {
            nameIfTrue: name @include(if: true)
            nameIfFalse: name @include(if: false)
        }
        """,
    )


def test_different_skip_include_directives_accepted(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment differentDirectivesWithDifferentAliases on Dog {
            name @include(if: true)
            name @include(if: false)
        }
        """,
    )


def test_same_aliases_with_different_field_targets(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment sameAliasesWithDifferentFieldTargets on Dog {
            fido: name
            fido: nickname
        }
        """,
        [
            'Field(s) "fido" conflict because "name" and "nickname" are '
            "different fields. Use different aliases on the fields to fetch "
            "both if this was intentional."
        ],
        [[(76, 86), (99, 113)]],
    )


def test_same_aliases_allowed_on_non_overlapping_fields(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment sameAliasesWithDifferentFieldTargets on Pet {
            ... on Dog {
                name
            }
            ... on Cat {
                name: nickname
            }
        }
        """,
    )


def test_alias_masking_direct_field_access(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment aliasMaskingDirectFieldAccess on Dog {
            name: nickname
            name
        }
        """,
        [
            'Field(s) "name" conflict because "nickname" and "name" are different '
            "fields. Use different aliases on the fields to fetch both if this "
            "was intentional."
        ],
        [[(69, 83), (96, 100)]],
    )


def test_different_args_second_adds_an_argument(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment conflictingArgs on Dog {
            doesKnowCommand
            doesKnowCommand(dogCommand: HEEL)
        }
        """,
        [
            'Field(s) "doesKnowCommand" conflict because "doesKnowCommand" '
            'and "doesKnowCommand" have different arguments. Use different '
            "aliases on the fields to fetch both if this was intentional."
        ],
    )


def test_different_args_second_missing_an_argument(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment conflictingArgs on Dog {
            doesKnowCommand(dogCommand: SIT)
            doesKnowCommand
        }
        """,
        [
            'Field(s) "doesKnowCommand" conflict because "doesKnowCommand" '
            'and "doesKnowCommand" have different arguments. Use different '
            "aliases on the fields to fetch both if this was intentional."
        ],
    )


def test_conflicting_args(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment conflictingArgs on Dog {
            doesKnowCommand(dogCommand: SIT)
            doesKnowCommand(dogCommand: HEEL)
        }
        """,
        [
            'Field(s) "doesKnowCommand" conflict because "doesKnowCommand" '
            'and "doesKnowCommand" have different arguments. Use different '
            "aliases on the fields to fetch both if this was intentional."
        ],
    )


def test_allows_different_args_where_no_conflict_is_possible(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment conflictingArgs on Pet {
            ... on Dog {
                name(surname: true)
            }
            ... on Cat {
                name
            }
        }
        """,
    )


def test_encounters_conflict_in_fragments(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            ...A
            ...B
        }
        fragment A on Type {
            x: a
        }
        fragment B on Type {
            x: b
        }
        """,
        [
            'Field(s) "x" conflict because "a" and "b" are different fields. Use '
            "different aliases on the fields to fetch both if this was intentional."
        ],
        [[(96, 100), (152, 156)]],
    )


def test_reports_each_conflict_once(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            f1 {
                ...A
                ...B
            }
            f2 {
                ...B
                ...A
            }
            f3 {
                ...A
                ...B
                x: c
            }
        }
        fragment A on Type {
            x: a
        }
        fragment B on Type {
            x: b
        }
        """,
        [
            'Field(s) "x" conflict because "a" and "b" are different fields. '
            "Use different aliases on the fields to fetch both if this was "
            "intentional.",
            'Field(s) "x" conflict because "c" and "a" are different fields. '
            "Use different aliases on the fields to fetch both if this was "
            "intentional.",
            'Field(s) "x" conflict because "c" and "b" are different fields. '
            "Use different aliases on the fields to fetch both if this was "
            "intentional.",
        ],
        [
            [(302, 306), (358, 362)],
            [(232, 236), (302, 306)],
            [(232, 236), (358, 362)],
        ],
    )


def test_deep_conflict(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            field {
                x: a
            },
            field {
                x: b
            }
        }
        """,
        [
            'Field(s) "field" conflict because subfields "x" conflict '
            '("a" and "b" are different fields). Use different aliases '
            "on the fields to fetch both if this was intentional."
        ],
        [[(23, 65), (47, 51), (79, 121), (103, 107)]],
    )


def test_deep_conflict_with_multiple_issues(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            field {
                x: a
                y: c
            },
            field {
                x: b
                y: d
            }
        }
        """,
        [
            'Field(s) "field" conflict because subfields "x" conflict '
            '("a" and "b" are different fields) and subfields "y" conflict '
            '("c" and "d" are different fields). Use different aliases on the '
            "fields to fetch both if this was intentional."
        ],
        [[(23, 86), (47, 51), (68, 72), (100, 163), (124, 128), (145, 149)]],
    )


def test_very_deep_conflict(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            field {
                deepField {
                    x: a
                }
            },
            field {
                deepField {
                    x: b
                }
            }
        }
        """,
        [
            'Field(s) "field" conflict because subfields "deepField" conflict '
            '(subfields "x" conflict ("a" and "b" are different fields)). '
            "Use different aliases on the fields to fetch both if this was "
            "intentional."
        ],
        [[(23, 115), (47, 101), (79, 83), (129, 221), (153, 207), (185, 189)]],
    )


def test_reports_deep_conflict_to_nearest_common_ancestor(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            field {
                deepField {
                    x: a
                }
                deepField {
                    x: b
                }
            },
            field {
                deepField {
                    y
                }
            }
        }
        """,
        [
            'Field(s) "deepField" conflict because subfields "x" conflict '
            '("a" and "b" are different fields). Use different aliases on the '
            "fields to fetch both if this was intentional."
        ],
        [[(47, 101), (79, 83), (118, 172), (150, 154)]],
    )


def test_reports_deep_conflict_to_nearest_common_ancestor_in_fragments(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            field {
                ...F
            }
            field {
                ...F
            }
        }
        fragment F on T {
            deepField {
                deeperField {
                    x: a
                }
                deeperField {
                    x: b
                }
            },
            deepField {
                deeperField {
                    y
                }
            }
        }
        """,
        [
            'Field(s) "deeperField" conflict because subfields "x" conflict '
            '("a" and "b" are different fields). Use different aliases on the '
            "fields to fetch both if this was intentional."
        ],
        [[(197, 253), (231, 235), (270, 326), (304, 308)]],
    )


def test_reports_deep_conflict_in_nested_fragments(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            field {
                ...F
            }
            field {
                ...I
            }
        }
        fragment F on T {
            x: a
            ...G
        }
        fragment G on T {
            y: c
        }
        fragment I on T {
            y: d
            ...J
        }
        fragment J on T {
            x: b
        }
        """,
        [
            'Field(s) "field" conflict because subfields "y" conflict ("c" and '
            '"d" are different fields) and subfields "x" conflict ("a" and "b" '
            "are different fields). "
            "Use different aliases on the fields to fetch both if this was "
            "intentional."
        ],
    )


def test_ignores_unknown_fragments(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
        field
        ...Unknown
        ...Known
        }

        fragment Known on T {
        field
        ...OtherUnknown
        }
        """,
    )


def test_conflicting_return_types_which_potentially_overlap(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ...on IntBox {
                    scalar
                }
                ...on NonNullStringBox1 {
                    scalar
                }
            }
        }
        """,
        [
            'Field(s) "scalar" conflict because they return '
            "conflicting types Int and String!. Use different aliases on the "
            "fields to fetch both if this was intentional."
        ],
    )


def test_compatible_return_shapes_on_different_return_types(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on SomeBox {
                deepBox {
                    unrelatedField
                }
                }
                ... on StringBox {
                    deepBox {
                        unrelatedField
                    }
                }
            }
        }
        """,
    )


def test_disallows_differing_return_types_despite_no_overlap(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    scalar
                }
                ... on StringBox {
                    scalar
                }
            }
        }
        """,
        [
            'Field(s) "scalar" conflict because they return conflicting types '
            "Int and String. Use different aliases on the fields to fetch both "
            "if this was intentional."
        ],
    )


def test_reports_correctly_when_a_non_exclusive_follows_an_exclusive(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    deepBox {
                    ...X
                    }
                }
            }
            someBox {
                ... on StringBox {
                    deepBox {
                    ...Y
                    }
                }
            }
            memoed: someBox {
                ... on IntBox {
                    deepBox {
                    ...X
                    }
                }
            }
            memoed: someBox {
                ... on StringBox {
                    deepBox {
                    ...Y
                    }
                }
            }
            other: someBox {
                ...X
            }
            other: someBox {
                ...Y
            }
        }
        fragment X on SomeBox {
            scalar
        }
        fragment Y on SomeBox {
            scalar: unrelatedField
        }
        """,
        [
            'Field(s) "other" conflict because subfields "scalar" conflict '
            '("scalar" and "unrelatedField" are different fields). Use different '
            "aliases on the fields to fetch both if this was intentional."
        ],
        [[(697, 748), (867, 873), (761, 812), (928, 950)]],
    )


def test_disallows_differing_return_type_nullability_despite_no_overlap(
    schema_2
):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on NonNullStringBox1 {
                    scalar
                }
                ... on StringBox {
                    scalar
                }
            }
        }
        """,
        [
            'Field(s) "scalar" conflict because they return conflicting types '
            "String! and String. Use different aliases on the fields to fetch "
            "both if this was intentional."
        ],
        [[(96, 102), (176, 182)]],
    )


def test_disallows_differing_return_type_list_despite_no_overlap_0(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    box: listStringBox {
                        scalar
                    }
                }
                ... on StringBox {
                    box: stringBox {
                    scalar
                    }
                }
            }
        }
        """,
        [
            'Field(s) "box" conflict because they return conflicting types '
            "[StringBox] and StringBox. Use different aliases on the fields to "
            "fetch both if this was intentional."
        ],
    )


def test_disallows_differing_return_type_list_despite_no_overlap_1(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    box: stringBox {
                    scalar
                    }
                }
                ... on StringBox {
                    box: listStringBox {
                    scalar
                    }
                }
            }
        }
        """,
        [
            'Field(s) "box" conflict because they return conflicting types '
            "StringBox and [StringBox]. Use different aliases on the fields to "
            "fetch both if this was intentional."
        ],
    )


def test_disallows_differing_subfields(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    box: stringBox {
                        val: scalar
                        val: unrelatedField
                    }
                }
                ... on StringBox {
                    box: stringBox {
                    val: scalar
                    }
                }
            }
        }
        """,
        [
            'Field(s) "val" conflict because "scalar" and "unrelatedField" are '
            "different fields. Use different aliases on the fields to fetch both "
            "if this was intentional."
        ],
    )


def test_disallows_differing_deep_return_types_despite_no_overlap(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    box: stringBox {
                    scalar
                    }
                }
                ... on StringBox {
                    box: intBox {
                    scalar
                    }
                }
            }
        }
        """,
        [
            'Field(s) "box" conflict because subfields "scalar" conflict '
            "(they return conflicting types String and Int). Use different "
            "aliases on the fields to fetch both if this was intentional."
        ],
    )


def test_allows_non_conflicting_overlaping_types(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ... on IntBox {
                    scalar: unrelatedField
                }
                ... on StringBox {
                    scalar
                }
            }
        }
        """,
    )


def test_same_wrapped_scalar_return_types(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ...on NonNullStringBox1 {
                    scalar
                }
                ...on NonNullStringBox2 {
                    scalar
                }
            }
        }
        """,
    )


def test_allows_inline_typeless_fragments(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        {
            a
            ... {
                a
            }
        }
        """,
    )


def test_compares_deep_types_including_list(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            connection {
                ...edgeID
                edges {
                    node {
                        id: name
                    }
                }
            }
        }

        fragment edgeID on Connection {
            edges {
                node {
                    id
                }
            }
        }
        """,
        [
            'Field(s) "edges" conflict because subfields "node" conflict '
            '(subfields "id" conflict ("name" and "id" are different fields)). '
            "Use different aliases on the fields to fetch both if this was "
            "intentional."
        ],
    )


def test_ignores_unknown_types(schema_2):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema_2,
        """
        {
            someBox {
                ...on UnknownType {
                    scalar
                }
                ...on NonNullStringBox2 {
                    scalar
                }
            }
        }
        """,
    )


def test_does_not_infinite_loop_on_recursive_fragment(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment fragA on Human { name, relatives { name, ...fragA } }
        """,
    )


def test_does_not_infinite_loop_on_immediately_recursive_fragment(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment fragA on Human { name, ...fragA }
        """,
    )


def test_does_not_infinite_loop_on_transitively_recursive_fragment(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment fragA on Human { name, ...fragB }
        fragment fragB on Human { name, ...fragC }
        fragment fragC on Human { name, ...fragA }
        """,
    )


def test_finds_invalid_case_even_with_immediately_recursive_fragment(schema):
    run_test(
        OverlappingFieldsCanBeMergedChecker,
        schema,
        """
        fragment sameAliasesWithDifferentFieldTargets on Dog {
            ...sameAliasesWithDifferentFieldTargets
            fido: name
            fido: nickname
        }
        """,
        [
            'Field(s) "fido" conflict because "name" and "nickname" are different '
            "fields. Use different aliases on the fields to fetch both if this "
            "was intentional."
        ],
    )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import NoFragmentCyclesChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_single_reference_is_valid(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB }
    fragment fragB on Dog { name }
    """,
    )


def test_spreading_twice_is_not_circular(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB, ...fragB }
    fragment fragB on Dog { name }
    """,
    )


def test_spreading_twice_indirectly_is_not_circular(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB, ...fragC }
    fragment fragB on Dog { ...fragC }
    fragment fragC on Dog { name }
    """,
        [],
    )


def test_double_spread_within_abstract_types(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment nameFragment on Pet {
        ... on Dog { name }
        ... on Cat { name }
    }

    fragment spreadsInAnon on Pet {
        ... on Dog { ...nameFragment }
        ... on Cat { ...nameFragment }
    }
    """,
        [],
    )


def test_does_not_false_positive_on_unknown_fragment(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment nameFragment on Pet {
        ...UnknownFragment
    }
    """,
        [],
    )


def test_spreading_recursively_within_field_fails(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Human { relatives { ...fragA } },
    """,
        ['Cannot spread fragment "fragA" withing itself'],
    )


def test_no_spreading_itself_directly(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragA }
    """,
        ['Cannot spread fragment "fragA" withing itself'],
    )


def test_no_spreading_itself_directly_within_inline_fragment(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Pet {
        ... on Dog {
            ...fragA
        }
    }
    """,
        ['Cannot spread fragment "fragA" withing itself'],
    )


def test_no_spreading_itself_indirectly(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB }
    fragment fragB on Dog { ...fragA }
    """,
        ['Cannot spread fragment "fragA" withing itself (via: fragB)'],
    )


def test_no_spreading_itself_indirectly_reports_opposite_order(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragB on Dog { ...fragA }
    fragment fragA on Dog { ...fragB }
    """,
        ['Cannot spread fragment "fragB" withing itself (via: fragA)'],
    )


def test_no_spreading_itself_indirectly_within_inline_fragment(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Pet {
        ... on Dog {
            ...fragB
        }
    }
    fragment fragB on Pet {
        ... on Dog {
            ...fragA
        }
    }
    """,
        ['Cannot spread fragment "fragA" withing itself (via: fragB)'],
    )


def test_no_spreading_itself_deeply(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB }
    fragment fragB on Dog { ...fragC }
    fragment fragC on Dog { ...fragO }
    fragment fragX on Dog { ...fragY }
    fragment fragY on Dog { ...fragZ }
    fragment fragZ on Dog { ...fragO }
    fragment fragO on Dog { ...fragP }
    fragment fragP on Dog { ...fragA, ...fragX }
    """,
        [
            'Cannot spread fragment "fragA" withing itself (via: fragB > fragC '
            "> fragO > fragP)",
            # Same priority but different iteration order from ref implementation.
            # 'Cannot spread fragment "fragO" withing itself (via: fragP > fragX '
            # '> fragY > fragZ)',
            'Cannot spread fragment "fragX" withing itself (via: fragY > fragZ '
            "> fragO > fragP)",
        ],
    )


def test_no_spreading_itself_deeply_two_paths(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB, ...fragC }
    fragment fragB on Dog { ...fragA }
    fragment fragC on Dog { ...fragA }
    """,
        [
            'Cannot spread fragment "fragA" withing itself (via: fragB)',
            # 'Cannot spread fragment "fragA" withing itself (via: fragC)',
        ],
    )


def test_no_spreading_itself_deeply_two_paths_alt_traverse_order(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragC }
    fragment fragB on Dog { ...fragC }
    fragment fragC on Dog { ...fragA, ...fragB }
    """,
        [
            'Cannot spread fragment "fragA" withing itself (via: fragC)',
            'Cannot spread fragment "fragB" withing itself (via: fragC)',
        ],
    )


def test_no_spreading_itself_deeply_and_immediately(schema):
    run_test(
        NoFragmentCyclesChecker,
        schema,
        """
    fragment fragA on Dog { ...fragB }
    fragment fragB on Dog { ...fragB, ...fragC }
    fragment fragC on Dog { ...fragA, ...fragB }
    """,
        [
            'Cannot spread fragment "fragB" withing itself',
            'Cannot spread fragment "fragA" withing itself (via: fragB > fragC)',
            # 'Cannot spread fragment "fragB" withing itself (via: fragC)',
        ],
    )



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

from py_gql.validation.rules import FieldsOnCorrectTypeChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_object_field_selection(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment objectFieldSelection on Dog {
            __typename
            name
        }""",
    )


def test_aliased_object_field_selection(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment aliasedObjectFieldSelection on Dog {
            tn : __typename
            otherName : name
        }
        """,
    )


def test_interface_field_selection(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment interfaceFieldSelection on Pet {
            __typename
            name
        }
        """,
    )


def test_aliased_interface_field_selection(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment interfaceFieldSelection on Pet {
            otherName : name
        }
        """,
    )


def test_lying_alias_selection(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment lyingAliasSelection on Dog {
            name : nickname
        }
        """,
    )


def test_ignores_fields_on_unknown_type(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment unknownSelection on UnknownType {
            unknownField
        }
        """,
    )


def test_reports_errors_when_type_is_known_again(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment typeKnownAgain on Pet {
            unknown_pet_field {
                ... on Cat {
                    unknown_cat_field
                }
            }
        }
        """,
        [
            'Cannot query field "unknown_pet_field" on type "Pet"',
            'Cannot query field "unknown_cat_field" on type "Cat"',
        ],
    )


def test_field_not_defined_on_fragment(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment fieldNotDefined on Dog {
            meowVolume
        }
        """,
        [
            'Cannot query field "meowVolume" on type "Dog", '
            'did you mean "barkVolume"?'
        ],
    )


def test_ignores_deeply_unknown_field(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment deepFieldNotDefined on Dog {
            unknown_field {
                deeper_unknown_field
            }
        }
        """,
        ['Cannot query field "unknown_field" on type "Dog"'],
    )


def test_sub_field_not_defined(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment subFieldNotDefined on Human {
            pets {
                unknown_field
            }
        }
        """,
        ['Cannot query field "unknown_field" on type "Pet"'],
    )


def test_field_not_defined_on_inline_fragment(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment fieldNotDefined on Pet {
            ... on Dog {
                meowVolume
            }
        }
        """,
        [
            'Cannot query field "meowVolume" on type "Dog", '
            'did you mean "barkVolume"?'
        ],
    )


def test_aliased_field_target_not_defined(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment aliasedFieldTargetNotDefined on Dog {
            volume : mooVolume
        }
        """,
        [
            'Cannot query field "mooVolume" on type "Dog", '
            'did you mean "barkVolume"?'
        ],
    )


def test_aliased_lying_field_target_not_defined(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment aliasedLyingFieldTargetNotDefined on Dog {
            barkVolume : kawVolume
        }
        """,
        [
            'Cannot query field "kawVolume" on type "Dog", '
            'did you mean "barkVolume"?'
        ],
    )


def test_not_defined_on_interface(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment notDefinedOnInterface on Pet {
            tailLength
        }
        """,
        ['Cannot query field "tailLength" on type "Pet"'],
    )


def test_defined_on_implementors_but_not_on_interface(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment definedOnImplementorsButNotInterface on Pet {
            nickname
        }
        """,
        ['Cannot query field "nickname" on type "Pet", did you mean "name"?'],
    )


def test_meta_field_selection_on_union(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment directFieldSelectionOnUnion on CatOrDog {
            __typename
        }
        """,
    )


def test_direct_field_selection_on_union(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment directFieldSelectionOnUnion on CatOrDog {
            directField
        }
        """,
        [
            'Cannot query field "directField" on type "CatOrDog", '
            'did you mean to use an inline fragment on "Dog" or "Cat"?'
        ],
    )


def test_defined_on_implementors_queried_on_union(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment definedOnImplementorsQueriedOnUnion on CatOrDog {
            name
        }
        """,
        [
            'Cannot query field "name" on type "CatOrDog", '
            'did you mean to use an inline fragment on "Dog" or "Cat"?'
        ],
    )


def test_valid_field_in_inline_fragment(schema):
    run_test(
        FieldsOnCorrectTypeChecker,
        schema,
        """
        fragment objectFieldSelection on Pet {
            ... on Dog {
                name
            }
            ... {
                name
            }
        }
        """,
    )



# -*- coding: utf-8 -*-
""" execution tests related to list types handling and resolution """

import pytest

from py_gql._utils import deduplicate
from py_gql.schema import (
    Field,
    Int,
    ListType,
    NonNullType,
    ObjectType,
    Schema,
    String,
)

from ._test_utils import check_execution


def run_test(
    test_type,
    test_data,
    expected_data=None,
    expected_errors=None,
    expected_exc=None,
    expected_msg=None,
):

    data = {"test": test_data}
    data_type = ObjectType(
        "DataType",
        [
            Field("test", test_type),
            Field("nest", lambda: data_type, resolve=lambda *_: data),
        ],
    )
    schema = Schema(data_type)

    check_execution(
        schema,
        "{ nest { test } }",
        initial_value=data,
        expected_data={"nest": {"test": expected_data}}
        if expected_data is not None
        else None,
        expected_errors=expected_errors,
        expected_exc=expected_exc,
        expected_msg=expected_msg,
    )


def _generator(iterable):
    for entry in iterable:
        yield entry


# Python sets are not ordered so here's a dumb iterable implementation
def _sortedset(values):
    return list(deduplicate(values))


def _lazy(values):
    return lambda *a, **kw: values


_FRUITS = ["apple", "banana", "apple", "coconut"]


@pytest.mark.parametrize(
    "iterable,result",
    [
        (_FRUITS, _FRUITS),
        (_sortedset(_FRUITS), ["apple", "banana", "coconut"]),
        (_generator(_FRUITS), _FRUITS),
    ],
)
def test_it_accepts_iterables_for_list_type(iterable, result):
    run_test(ListType(String), iterable, result, [])


@pytest.mark.parametrize("not_iterable", ["apple", 42, object()])
def test_it_raises_on_non_iterable_value_for_list_type(not_iterable):
    run_test(
        ListType(String),
        not_iterable,
        expected_exc=RuntimeError,
        expected_msg=(
            'Field "nest.test" is a list type and resolved value should '
            "be iterable"
        ),
    )


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param([1, 2], [1, 2], id="[T]"),
        pytest.param([1, None, 2], [1, None, 2], id="[T], contains null"),
        pytest.param(_lazy([1, 2]), [1, 2], id="[T], callable"),
        pytest.param(
            _lazy([1, None, 2]), [1, None, 2], id="[T], callable, contains null"
        ),
        pytest.param(None, None, id="[T], null"),
        pytest.param(_lazy(None), None, id="[T], callable, null"),
    ],
)
def test_nullable_list_of_nullable_items(data, expected):
    run_test(ListType(Int), data, expected)


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param([1, 2], [1, 2], id="[T]!"),
        pytest.param([1, None, 2], [1, None, 2], id="[T]!, contains null"),
        pytest.param(_lazy([1, 2]), [1, 2], id="[T]!, callable"),
        pytest.param(
            _lazy([1, None, 2]),
            [1, None, 2],
            id="[T]!, callable, contains null",
        ),
    ],
)
def test_non_nullable_list_of_nullable_items_ok(data, expected):
    run_test(NonNullType(ListType(Int)), data, expected)


@pytest.mark.parametrize(
    "data, expected_err",
    [
        (None, ('Field "nest.test" is not nullable', (9, 13), "nest.test")),
        (
            _lazy(None),
            ('Field "nest.test" is not nullable', (9, 13), "nest.test"),
        ),
    ],
)
def test_non_nullable_list_of_nullable_items_fail(data, expected_err):
    run_test(NonNullType(ListType(Int)), data, None, [expected_err])


@pytest.mark.parametrize(
    "data, expected",
    [
        pytest.param([1, 2], [1, 2], id="[T!]"),
        pytest.param(_lazy([1, 2]), [1, 2], id="[T!], callable"),
        pytest.param(None, None, id="[T!], null"),
        pytest.param(_lazy(None), None, id="[T!], callable, null"),
    ],
)
def test_nullable_list_of_non_nullable_items_ok(data, expected):
    run_test(ListType(NonNullType(Int)), data, expected)


@pytest.mark.parametrize(
    "data, expected_err",
    [
        (
            [1, None, 2],
            ('Field "nest.test[1]" is not nullable', (9, 13), "nest.test[1]"),
        ),
        (
            _lazy([1, None, 2]),
            ('Field "nest.test[1]" is not nullable', (9, 13), "nest.test[1]"),
        ),
    ],
)
def test_nullable_list_of_non_nullable_items_fail(data, expected_err):
    run_test(ListType(NonNullType(Int)), data, None, [expected_err])



# -*- coding: utf-8 -*-
""" Work with strings """
from __future__ import division

import operator
import re

import six

LINE_SEPARATOR = re.compile(r"\r\n|[\n\r]")
LEADING_WS = re.compile(r"^[\t\s]*")


def ensure_unicode(string):
    if isinstance(string, six.binary_type):
        return string.decode("utf8")
    return string


def leading_whitespace(string):
    r""" Detect of leading whitespace in a string.

    Args:
        string (str): Input value

    Returns:
        int: Length of leading whitespace

    >>> leading_whitespace('  \t  foo')
    5
    >>> leading_whitespace('\tfoo')
    1
    >>> leading_whitespace('   foo')
    3
    """
    return len(string) - len(LEADING_WS.sub("", string))


def is_blank(string):
    """
    Args:
        string (str): Input value

    Returns:
        bool: Whether the string is blank
    """
    return not string.strip()


def parse_block_string(
    raw_string, strip_trailing_newlines=True, strip_leading_newlines=True
):
    """ Parse a raw string according to the GraphQL spec's BlockStringValue()
    http://facebook.github.io/graphql/draft/#BlockStringValue() static
    algorithm. Similar to Coffeescript's block string, Python's docstring trim
    or Ruby's strip_heredoc.

    Args:
        raw_string (str): Input value
        strip_trailing_newlines (bool): Remove trailing newlines
        strip_leading_newlines (bool): Remove leading newlines

    Returns:
        str: Block string value
    """
    lines = LINE_SEPARATOR.split(raw_string)
    common_indent = None

    for i, line in enumerate(lines):
        if i == 0:
            continue
        indent = leading_whitespace(line)
        if indent < len(line) and (
            common_indent is None or indent < common_indent
        ):
            common_indent = indent

    if common_indent:
        lines = [
            line[common_indent:]
            if (len(line) >= common_indent and i > 0)
            else line
            for i, line in enumerate(lines)
        ]

    while strip_leading_newlines and lines and is_blank(lines[0]):
        lines.pop(0)

    while strip_trailing_newlines and lines and is_blank(lines[-1]):
        lines.pop()

    return "\n".join(lines)


dedent = lambda s: parse_block_string(s, strip_trailing_newlines=False)


def index_to_loc(body, position):
    r""" Get the (line number, column number) tuple from a zero-indexed offset.

    Args:
        body (str): Source string
        position (int): 0-indexed position of the character

    Returns:
        Tuple[int, int]: (line number, column number)

    Raises:
        :py:class:`IndexError`: if ``position`` is out of bounds

    >>> index_to_loc("ab\ncd\ne", 0)
    (1, 1)

    >>> index_to_loc("ab\ncd\ne", 3)
    (2, 1)

    >>> index_to_loc("", 0)
    (1, 1)

    >>> index_to_loc("{", 1)
    (1, 2)

    >>> index_to_loc("", 42)
    Traceback (most recent call last):
        ...
    IndexError: 42
    """
    if not body and not position:
        return (1, 1)

    if position > len(body) or position < 0:
        raise IndexError(position)

    lines, cols = 0, 0
    for offset, char in enumerate(body):
        if offset == position:
            return (lines + 1, cols + 1)
        elif char == "\n":
            lines += 1
            cols = 0
        else:
            cols += 1
    return (lines + 1, cols + 1)


def loc_to_index(body, loc):
    r""" Get the zero-indexed offset from a (lineno, col) tuple.

    Args:
        body (str): Source string
        loc (Tuple[int, int]): (line number, column number)

    Returns:
        int: 0-indexed position of the character

    Raises:
        :py:class:`IndexError`: if ``loc`` is out of bounds

    >>> loc_to_index("ab\ncd\ne", (1, 1))
    0

    >>> loc_to_index("", (1, 1))
    0

    >>> loc_to_index("ab\ncd\ne", (2, 1))
    3

    >>> loc_to_index("{", (1, 2))
    1

    >>> loc_to_index("ab\ncd\ne", (6, 7))
    Traceback (most recent call last):
        ...
    IndexError: 6:7
    """
    if not body and loc == (1, 1):
        return 0

    lineo, col = loc
    lines = 0
    for index, char in enumerate(body):
        if lines == lineo - 1:
            if len(body) >= index + col - 1:
                return index + col - 1
            break
        if char == "\n":
            lines += 1
    raise IndexError("%s:%s" % (lineo, col))


def highlight_location(body, position, delta=2):
    """ Nicely format a highlited view of a position into a source string.

    Args:
        body (str): Source string
        position (int): 0-indexed position of the character
        delta (int): How many lines around the position should this conserve

    Returns:
        str: Formatted view

    >>> print(highlight_location('''{
    ...     query {
    ...         Node (search: "foo") {
    ...             id
    ...             name
    ...         }
    ...     }
    ... }''', 40))
    (3:27):
      1:{
      2:    query {
      3:        Node (search: "foo") {
                                  ^
      4:            id
      5:            name
    """
    # REVIEW: There must be a more readable way to write this
    line, col = index_to_loc(body, position)
    line_index = line - 1
    lines = LINE_SEPARATOR.split(body)
    min_line = max(0, line_index - delta)
    max_line = min(line_index + delta, len(lines) - 1)
    pad_len = len(str(max_line + 1))

    def ws(len_):
        return "".join((" " for _ in range(len_)))

    def lineno(l):
        return " ".join(range(pad_len - len(str(l + 1)))) + str(l + 1)

    output = ["(%d:%d):" % (line, col)]
    output.extend(
        [
            ws(2) + lineno(l) + ":" + lines[l]
            for l in range(min_line, line_index)
        ]
    )
    output.append(ws(2) + lineno(line_index) + ":" + lines[line_index])
    output.append(ws(2) + ws(len(str(max_line + 1)) + col) + "^")
    output.extend(
        [
            ws(2) + lineno(l) + ":" + lines[l]
            for l in range(line_index + 1, max_line + 1)
        ]
    )
    return "\n".join(output)


def _split_words_with_boundaries(string, word_boundaries):
    """
    >>> list(_split_words_with_boundaries("ab cd -ef_gh", " -_"))
    ['ab', ' ', 'cd', ' ', '-', 'ef', '_', 'gh']
    """
    stack = []
    for char in string:
        if char in word_boundaries:
            if stack:
                yield "".join(stack)
            yield char
            stack[:] = []
        else:
            stack.append(char)

    if stack:
        yield "".join(stack)


def wrapped_lines(lines, max_len, word_boundaries=" -_"):
    """ Wrap provided lines to a given length.

    Lines that are under ``max_len`` are left as is, otherwise this splits
    them based on the specified word boundaries and yields parts of the line
    that under ``max_len`` until the line has been exhausted,

    Args:
        lines (Iterator[str]): Source lines
        max_len (int): Maximum line length
        word_boundaries (Iterator[str]): Which charcaters are used to split lines

    Yields:
        str: The next wrapped line
    """
    for line in lines:
        if len(line) <= max_len:
            yield line
            continue

        wrapped = ""

        for entry in _split_words_with_boundaries(line, word_boundaries):
            if len(wrapped + entry) > max_len:
                yield wrapped
                wrapped = ""
            if entry != " " or wrapped:
                wrapped += entry

        if wrapped:
            yield wrapped


def levenshtein(s1, s2):
    """ Compute the Levenshtein edit distance between 2 strings.

    Args:
        s1 (str): First string
        s2 (str): Second string

    Returns:
        int: Computed edit distance
    """
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def infer_suggestions(candidate, options, distance=levenshtein):
    """ Extract the most similar entries to an input string given multiple
    options and a distance function.

    Args:
        candidate (str): Input string
        options (Iterator[str]): Possible options
        distance (callable):
            Distance function, must have the signature ``(s1, s2) -> int``
            where the more similar the inputs are, the lower the result is.

    Returns:
        List[str]: Most similar options sorted by similarity (most similar to
        least similar)
    """
    distances = []
    half = len(candidate) / 2
    for option in options:
        distance = levenshtein(candidate, option)
        threshold = max(half, len(option) / 2, 1)
        if distance <= threshold:
            distances.append((option, distance))
    return [opt for opt, _ in sorted(distances, key=operator.itemgetter(1))]


def quoted_options_list(options):
    """ Quote a list of possible strings.

    Args:
        options (Iterator[str]): Possible options

    Returns:
        str: Quoted options

    >>> quoted_options_list([])
    ''

    >>> quoted_options_list(['foo'])
    '"foo"'

    >>> quoted_options_list(['foo', 'bar'])
    '"foo" or "bar"'

    >>> quoted_options_list(['foo', 'bar', 'baz'])
    '"foo", "bar" or "baz"'
    """
    if not options:
        return ""

    if len(options) == 1:
        return '"%s"' % options[0]

    return "%s or %s" % (
        ", ".join(('"%s"' % option for option in options[:-1])),
        '"%s"' % options[-1],
    )


def stringify_path(path):
    """ Concatenate traversal path into a string.

    >>> stringify_path(['foo', 0, 'bar'])
    'foo[0].bar'

    >>> stringify_path(['foo'])
    'foo'

    >>> stringify_path([1, 'foo'])
    '[1].foo'

    >>> stringify_path([])
    ''
    """
    path_str = ""
    for entry in path:
        if isinstance(entry, int):
            path_str += "[%s]" % entry
        else:
            path_str += ".%s" % entry
    return path_str.lstrip(".")



# -*- coding: utf-8 -*-
""" Test specified rule in isolation. """

# Tests were adapted from the one in the GraphQLJS reference implementation,
# as our version exits early not all of the expected errors are aplicable but
# they conserved as comments for reference.
# Tests related to suggestion list are kept for reference but skipped as this
# feature is not implemented.

from py_gql.validation.rules import VariablesInAllowedPositionChecker

from .._test_utils import assert_checker_validation_result as run_test


def test_boolean_to_boolean(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($booleanArg: Boolean)
        {
            complicatedArgs {
                booleanArgField(booleanArg: $booleanArg)
            }
        }
        """,
    )


def test_boolean_to_boolean_within_fragment_0(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        fragment booleanArgFrag on ComplicatedArgs {
            booleanArgField(booleanArg: $booleanArg)
        }

        query Query($booleanArg: Boolean) {
            complicatedArgs {
                ...booleanArgFrag
            }
        }
        """,
    )


def test_boolean_to_boolean_within_fragment_1(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($booleanArg: Boolean) {
            complicatedArgs {
                ...booleanArgFrag
            }
        }

        fragment booleanArgFrag on ComplicatedArgs {
            booleanArgField(booleanArg: $booleanArg)
        }
        """,
    )


def test_required_boolean_to_boolean(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($nonNullBooleanArg: Boolean!) {
            complicatedArgs {
                booleanArgField(booleanArg: $nonNullBooleanArg)
            }
        }
        """,
    )


def test_required_boolean_to_boolean_within_fragment(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        fragment booleanArgFrag on ComplicatedArgs {
            booleanArgField(booleanArg: $nonNullBooleanArg)
        }

        query Query($nonNullBooleanArg: Boolean!) {
            complicatedArgs {
                ...booleanArgFrag
            }
        }
        """,
    )


def test_int_to_required_int_with_default(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($intArg: Int = 1)
        {
            complicatedArgs {
                nonNullIntArgField(nonNullIntArg: $intArg)
            }
        }
        """,
    )


def test_list_of_string_to_list_of_string(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringListVar: [String]) {
            complicatedArgs {
                stringListArgField(stringListArg: $stringListVar)
            }
        }
        """,
    )


def test_list_of_required_string_to_list_of_string(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringListVar: [String!]) {
            complicatedArgs {
                stringListArgField(stringListArg: $stringListVar)
            }
        }
        """,
    )


def test_string_to_list_of_string_in_field(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringVar: String) {
            complicatedArgs {
                stringListArgField(stringListArg: [$stringVar])
            }
        }
        """,
    )


def test_required_string_to_list_of_string(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringVar: String!) {
            complicatedArgs {
                stringListArgField(stringListArg: [$stringVar])
            }
        }
        """,
    )


def test_complex_input_to_complex_input(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($complexVar: ComplexInput) {
            complicatedArgs {
                complexArgField(complexArg: $complexVar)
            }
        }
        """,
    )


def test_complex_input_field_to_complex_input_field_0(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($boolVar: Boolean = false) {
            complicatedArgs {
                complexArgField(complexArg: {requiredField: $boolVar})
            }
        }
        """,
    )


def test_complex_input_field_to_complex_input_field_1(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($boolVar: Boolean = false) {
            complicatedArgs {
                complexArgField(complexArg: {requiredArg: $boolVar})
            }
        }
        """,
    )


def test_req_boolean_to_req_boolean_in_directive(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($boolVar: Boolean!) {
            dog @include(if: $boolVar)
        }
        """,
    )


def test_boolean_to_req_boolean_in_directive_with_default(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($boolVar: Boolean = false) {
            dog @include(if: $boolVar)
        }
        """,
    )


def test_int_to_required_int(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($intArg: Int) {
            complicatedArgs {
                nonNullIntArgField(nonNullIntArg: $intArg)
            }
        }
        """,
        ['Variable "$intArg" of type Int used in position expecting type Int!'],
        [(117, 124)],
    )


def test_int_to_required_int_within_fragment(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        fragment nonNullIntArgFieldFrag on ComplicatedArgs {
            nonNullIntArgField(nonNullIntArg: $intArg)
        }

        query Query($intArg: Int) {
            complicatedArgs {
                ...nonNullIntArgFieldFrag
            }
        }
        """,
        ['Variable "$intArg" of type Int used in position expecting type Int!'],
        [(108, 115)],
    )


def test_int_to_required_int_within_nested_fragment(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        fragment outerFrag on ComplicatedArgs {
            ...nonNullIntArgFieldFrag
        }

        fragment nonNullIntArgFieldFrag on ComplicatedArgs {
            nonNullIntArgField(nonNullIntArg: $intArg)
        }

        query Query($intArg: Int) {
            complicatedArgs {
                ...outerFrag
            }
        }
        """,
        ['Variable "$intArg" of type Int used in position expecting type Int!'],
        [(205, 212)],
    )


def test_string_over_boolean(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringVar: String) {
            complicatedArgs {
                booleanArgField(booleanArg: $stringVar)
            }
        }
        """,
        [
            'Variable "$stringVar" of type String used in position expecting '
            "type Boolean"
        ],
    )


def test_string_to_list_of_string(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringVar: String) {
            complicatedArgs {
                stringListArgField(stringListArg: $stringVar)
            }
        }
        """,
        [
            'Variable "$stringVar" of type String used in position expecting '
            "type [String]"
        ],
    )


def test_boolean_to_required_boolean_in_directive(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($boolVar: Boolean) {
            dog @include(if: $boolVar)
        }
        """,
        [
            'Variable "$boolVar" of type Boolean used in position expecting '
            "type Boolean!"
        ],
    )


def test_string_to_required_boolean_in_directive(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringVar: String) {
            dog @include(if: $stringVar)
        }
        """,
        [
            'Variable "$stringVar" of type String used in position expecting '
            "type Boolean!"
        ],
    )


def test_list_of_string_to_list_of_required_string(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($stringListVar: [String]) {
            complicatedArgs {
                stringListNonNullArgField(stringListNonNullArg: $stringListVar)
            }
        }
        """,
        [
            'Variable "$stringListVar" of type [String] used in position '
            "expecting type [String!]"
        ],
    )


def test_int_to_non_null_int_fails_when_variable_provides_null_default(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($intVar: Int = null) {
            complicatedArgs {
                nonNullIntArgField(nonNullIntArg: $intVar)
            }
        }
        """,
        ['Variable "$intVar" of type Int used in position expecting type Int!'],
        [(124, 131)],
    )


def test_int_to_non_null_int_ok_when_variable_provides_non_null_default(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($intVar: Int = 1) {
            complicatedArgs {
                nonNullIntArgField(nonNullIntArg: $intVar)
            }
        }
        """,
    )


def test_int_to_non_null_int_ok_when_optional_argument_provides_default(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($intVar: Int) {
            complicatedArgs {
                nonNullFieldWithDefault(nonNullIntArg: $intVar)
            }
        }
        """,
    )


def test_bool_to_non_null_bool_in_directive_with_default_with_option(schema):
    run_test(
        VariablesInAllowedPositionChecker,
        schema,
        """
        query Query($boolVar: Boolean = false) {
            dog @include(if: $boolVar)
        }
        """,
    )



# -*- coding: utf-8 -*-

import json
from concurrent.futures import Future

import six

from .._utils import OrderedDict
from ..exc import CoercionError, MultiCoercionError, ResolverError
from ..utilities import directive_arguments


class ExecutionContext(object):
    """ Container to be passed around **internally** during the execution.
    Unless you are implementing a custom execution function you should not need to
    refer to this class.
    """

    __slots__ = (
        "schema",
        "document",
        "variables",
        "fragments",
        "executor",
        "future_cls",
        "operation",
        "middlewares",
        "context",
        "_errors",
        "grouped_fields",
        "field_defs",
        "subselections",
        "argument_values",
    )

    def __init__(
        self,
        schema,
        document,
        variables,
        fragments,
        executor,
        operation,
        middlewares,
        context,
    ):
        self.schema = schema
        self.document = document
        self.variables = variables
        self.fragments = fragments
        self.executor = executor
        self.future_cls = getattr(executor, "Future", Future)
        self.operation = operation
        self.middlewares = middlewares
        self.context = context

        self._errors = []

        self.grouped_fields = {}
        self.field_defs = {}
        self.subselections = {}
        self.argument_values = {}

    def add_error(self, err, node=None, path=None):
        """ Register a localized execution error.

        :type err: str|Exception
        :param err: The error

        :type node: py_gql.lang.ast.Node
        :param node: The node corresponding to this error

        :type path: list
        :param path: The traversal path where this error was occuring
        """
        if isinstance(err, six.string_types):
            err = ResolverError(err, [node], path)
        elif isinstance(err, MultiCoercionError):
            for child_error in err.errors:
                self.add_error(child_error)
        elif isinstance(err, (ResolverError, CoercionError)):
            if node:
                if err.nodes and node not in err.nodes:
                    err.nodes.append(node)
                elif not err.nodes:
                    err.nodes = [node]
            err.path = path if path is not None else err.path
        self._errors.append((err, node, path))

    @property
    def errors(self):
        """ Get a copy of the errors without the risk of modifying the internal
        structure.

        :rtype: list[tuple[str|Exception, py_gql.lang.ast.Node, list]]
        """
        return self._errors[:]


# REVIEW: Maybe this exposes too much?
class ResolveInfo(object):
    """ ResolveInfo will be passed to resolver functions to expose
    some of the data used in the execution process as well as some common helper
    functions. You shouldn't instantiate thos yourself.

    Warning:

        Interface will surely change as the resolver definitions gets
        fined-tuned.

    Args:
        field_def (py_gql.schema.Field):
        parent_type (py_gql.schema.ObjectType):
        path (List[Union[str, int]]):
        schema (py_gql.schema.Schema):
        variables (dict):
        fragments (dict[str, py_gql.lang.ast.FragmentDefinition]):
        operation (py_gql.lang.ast.OperationDefinition):
        node (list[py_gql.lang.ast.Node]):
        executor (py_gql.execution.executors.Executor):

    Attributes:
        field_def (py_gql.schema.Field): Field being resolved

        parent_type (py_gql.schema.ObjectType):
            ObjectType definition where the field originated from

        path (List[Union[str, int]]): Current traversal path

        schema (py_gql.schema.Schema): Schema

        variables (dict): Coerced variables

        fragments (dict[str, py_gql.lang.ast.FragmentDefinition]):
            Fragments present in the document

        operation (py_gql.lang.ast.OperationDefinition): Current operation

        node (list[py_gql.lang.ast.Node]): AST nodes corresponding to the field

        executor (py_gql.execution.executors.Executor): Current executor
            Use this if you need to submit call other resover / fetcher
            functions inside a resolver.

    """

    __slots__ = (
        "field_def",
        "parent_type",
        "path",
        "schema",
        "variables",
        "fragments",
        "operation",
        "nodes",
        "_directive_values",
        "executor",
    )

    def __init__(
        self,
        field_def,
        parent_type,
        path,
        schema,
        variables,
        fragments,
        operation,
        nodes,
        executor,
    ):
        self.field_def = field_def
        self.parent_type = parent_type
        self.path = path
        self.schema = schema
        self.variables = variables
        self.fragments = fragments
        self.operation = operation
        self.nodes = nodes
        self.executor = executor

        self._directive_values = {}

    def directive_values(self, directive_name):
        """ Extract directive argument values for given directive name.

        :type directive_name: str
        :param directive_name: Directive name.
            This assumes validation has run and the directive exists in the schema
            as well as in a valid position.

        :rtype: dict
        :returns: The arguments to the directive
        """
        if directive_name in self._directive_values:
            return self._directive_values[directive_name]

        definition = self.schema.directives[directive_name]
        values = {}
        for node in self.nodes:
            values.update(
                directive_arguments(definition, node, self.variables) or {}
            )
        self._directive_values[directive_name] = values
        return values


_unset = object()


class GraphQLResult(object):
    """ Wrapper encoding the behaviour described in the `Response
    <http://facebook.github.io/graphql/June2018/#sec-Response>`_ part of the
    specification.

    Args:
        data (Optional[Any]):
            The data part of the response
        errors (Optional[List[py_gql.exc.GraphQLResponseError]]):
            The errors part of the response
    """

    __slots__ = "data", "errors", "extensions"

    def __init__(self, data=_unset, errors=_unset):
        self.data = data
        self.errors = errors
        self.extensions = OrderedDict()

    def __bool__(self):
        return self.errors is _unset or not self.errors

    __nonzero__ = __bool__

    def __iter__(self):
        return iter((self.data, self.errors))

    def add_extension(self, ext):
        """ Add an extensions to the result.

        Args:
            ext (GraphQLExtension): Extension instance

        Raises:
            ValueError: Extension with the same name has already been added
        """
        name = ext.name()
        if name in self.extensions:
            raise ValueError('Duplicate extension "%s"' % name)
        self.extensions[name] = ext.payload()

    def response(self):
        """ Generate an ordered response dict.

        Returns:
            dict:
        """
        d = OrderedDict()
        if self.errors is not _unset and self.errors:
            d["errors"] = [error.to_dict() for error in self.errors]
        if self.data is not _unset:
            d["data"] = self.data
        if self.extensions:
            d["extensions"] = self.extensions
        return d

    def json(self, **kw):
        """ Encode response as JSON using the standard lib ``json`` module.

        Args:
            **kw (dict): Keyword args passed to to ``json.dumps``

        Returns:
            str:
        """
        return json.dumps(self.response(), **kw)


class GraphQLExtension(object):
    """ Encode a GraphQL response extension.

    Use in conjonction with :meth:`GraphQLResult.add_extension` to encode the
    response alongside an execution result.
    """

    def payload(self):
        """
        Returns:
            Any: Extension payload; **must** be JSON serialisable.
        """
        raise NotImplementedError()

    def name(self):
        """
        Returns:
            str: Name of the extension used as the key in the response.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
""" """
from __future__ import print_function

from py_gql.lang import parse
from py_gql.validation import validate_ast


def _ensure_list(value):
    if isinstance(value, list):
        return value
    else:
        return [value]


def assert_validation_result(
    schema, source, expected_msgs=None, expected_locs=None, checkers=None
):
    # Prints are here so we can more easily debug when running pytest with -v
    expected_msgs = expected_msgs or []
    expected_locs = expected_locs or []

    print(source)
    result = validate_ast(schema, parse(source), checkers)
    errors = result.errors

    msgs = [str(err) for err in errors]
    locs = [[node.loc for node in err.nodes] for err in errors]

    print(" [msgs] ", msgs)
    print(" [locs] ", locs)

    assert msgs == expected_msgs
    if expected_locs:
        assert locs == [_ensure_list(l) for l in expected_locs]


def assert_checker_validation_result(
    checker, schema, source, expected_msgs=None, expected_locs=None
):
    assert_validation_result(
        schema,
        source,
        expected_msgs=expected_msgs,
        expected_locs=expected_locs,
        checkers=[checker],
    )



# -*- coding: utf-8 -*-
""" Iterable interface for the GraphQL Language lexer. """

# [TODO] Review the various `Lexer.read_*` methods as the use of `peek` and
# `advance` is fairly inconsistent from tweaking the code around to make the
# tests pass. They could certainly be streamlined / optimised.

# [TODO] Settle on using HEX codes or chars, this make the code a bit
# inconsistent.

import six

from . import token
from .._string_utils import ensure_unicode, parse_block_string
from ..exc import (
    InvalidCharacter,
    InvalidEscapeSequence,
    NonTerminatedString,
    UnexpectedCharacter,
    UnexpectedEOF,
)

EOL_CHARS = frozenset([0x000A, 0x000D])  # "\n"  # "\r"

IGNORED_CHARS = (
    frozenset([0xFEFF, 0x0009, 0x0020, 0x002C])
    | EOL_CHARS  # BOM  # \t  # SPACE  # ,
)

SYMBOLS = {
    cls.__char__: cls
    for cls in (
        token.ExclamationMark,
        token.Dollar,
        token.ParenOpen,
        token.ParenClose,
        token.BracketOpen,
        token.BracketClose,
        token.CurlyOpen,
        token.CurlyClose,
        token.Colon,
        token.Equals,
        token.At,
        token.Pipe,
        token.Ampersand,
    )
}

QUOTED_CHARS = {
    0x0022: u'"',
    0x005c: u"\\",
    0x002f: u"/",
    0x0062: u"\u0008",
    0x0066: u"\u000c",
    0x006e: u"\n",
    0x0072: u"\r",
    0x0074: u"\t",
}


def _unexpected(expected, char, position, source):
    if char is None:
        return UnexpectedEOF(position - 1, source)
    else:
        return UnexpectedCharacter(
            'Expected "%s" but found "%s"' % (expected, char), position, source
        )


class Lexer(object):
    """ Iterable GraphQL language lexer / tokenizer.

    This class is not typically exposed through the parser but can be used
    independently to build custom parsers.

    Each call to ``__next__`` will read over a number of characters required
    to form a valid :class:`py_gql.lang.token.Token` and otherwise raise
    :class:`~py_gql.exc.GraphQLSyntaxError` if that is not possible.

    Args:
        source (Union[str, bytes]): Source string.
            Bytestrings will be converted to unicode.
    """

    __slots__ = ("_source", "_len", "_done", "_position", "_started")

    def __init__(self, source):

        if source is None:
            raise ValueError("source cannot be None")

        self._source = ensure_unicode(source)
        self._len = len(source)
        self._done = False
        self._started = False
        self._position = 0

    def _peek(self):
        try:
            return self._source[self._position]
        except IndexError:
            return None

    def _range(self, start=0, end=1):
        return self._source[self._position + start : self._position + end]

    def _read_over_whitespace(self):
        """ Advance lexer over all ignored characters. """
        pos = self._position
        while True:
            try:
                char = self._source[pos]
            except IndexError:
                break

            code = ord(char)

            if code in IGNORED_CHARS:
                pos += 1
            elif code == 0x0023:  # '#'
                pos += 1
                while True:
                    try:
                        char = self._source[pos]
                    except IndexError:
                        break

                    code = ord(char)
                    if (
                        code >= 0x0020 or code == 0x0009
                    ) and code not in EOL_CHARS:
                        pos += 1
                    else:
                        break
            else:
                break

        self._position = pos

    def _read_ellipsis(self):
        """ Advance lexer over an ellipsis token (...).

        Returns:
            py_gql.lang.token.Ellipsis_: parse token
        """
        start = self._position
        for _ in range(3):
            char = self._peek()
            self._position += 1
            if char != ".":
                raise _unexpected(".", char, self._position, self._source)
        return token.Ellipsis_(start, self._position)

    def _read_string(self):
        """ Advance lexer over a quoted string.

        Returns:
            py_gql.lang.token.String: parse token
        """
        start = self._position
        self._position += 1
        acc = []
        while True:
            char = self._peek()

            if char is None:
                raise NonTerminatedString("", self._position, self._source)

            code = ord(char)
            self._position += 1

            if char == '"':
                value = "".join(acc)
                return token.String(start, self._position, value)
            elif char == "\\":
                acc.append(self._read_escape_sequence())
            elif code == 0x000a or code == 0x000d:  # \n or \r
                raise NonTerminatedString("", self._position - 1, self._source)
            elif not (code >= 0x0020 or code == 0x0009):
                raise InvalidCharacter(char, self._position - 1, self._source)
            else:
                acc.append(char)

        raise NonTerminatedString("", self._position, self._source)

    def _read_block_string(self):
        """ Advance lexer over a triple quoted block string.

        Returns:
            py_gql.lang.token.BlockString: parse token
        """
        start = self._position
        self._position += 3
        acc = []

        while True:
            char = self._peek()

            if char is None:
                raise NonTerminatedString("", self._position, self._source)

            code = ord(char)

            if self._range(0, 3) == '"""':
                self._position += 3
                value = parse_block_string("".join(acc))
                return token.BlockString(start, self._position, value)

            self._position += 1

            if char == "\\":
                if self._range(0, 3) == '"""':
                    for _ in range(3):
                        acc.append(self._peek())
                        self._position += 1
                else:
                    acc.append(char)
            elif not (code >= 0x0020 or code == 0x0009 or code in EOL_CHARS):
                raise InvalidCharacter(char, self._position - 1, self._source)
            else:
                acc.append(char)

        raise NonTerminatedString("", self._position, self._source)

    def _read_escape_sequence(self):
        """ Advance lexer over an escape character

        Returns:
            chr: Escaped character value
        """
        char = self._peek()
        self._position += 1
        if char is None:
            raise NonTerminatedString("", self._position, self._source)

        code = ord(char)

        if code in QUOTED_CHARS:
            return QUOTED_CHARS[code]
        elif code == 0x0075:  # unicode character: uXXXX
            return self._read_escaped_unicode()
        else:
            raise InvalidEscapeSequence(
                u"\\%s" % char, self._position - 1, self._source
            )

    def _read_escaped_unicode(self):
        """ Advance lexer over a unicode character

        Returns:
            chr: Escaped character value
        """
        start = self._position
        for _ in range(4):
            char = self._peek()
            self._position += 1
            if char is None:
                raise NonTerminatedString("", self._position, self._source)
            if not char.isalnum():
                break

        escape = self._source[start : self._position]

        if len(escape) != 4:
            raise InvalidEscapeSequence(
                u"\\u%s" % escape, start - 1, self._source
            )

        try:
            return u"%c" % six.unichr(int(escape, 16))
        except ValueError:
            raise InvalidEscapeSequence(
                u"\\u%s" % escape, start - 1, self._source
            )

    def _read_number(self):
        """ Advance lexer over a number

        Returns:
            Union[py_gql.lang.token.Integer, py_gql.lang.token.Float]:
                parse token
        """
        start = self._position
        is_float = False

        char = self._peek()
        if ord(char) == 0x002d:  # "-"
            self._position += 1

        self._read_over_integer()

        char = self._peek()

        if char is not None and ord(char) == 0x002e:  # "."
            self._position += 1
            is_float = True
            self._read_over_digits()

        char = self._peek()

        if char is not None and ord(char) in (0x0065, 0x0045):  # "e", "E"
            self._position += 1
            is_float = True
            char = self._peek()
            if char is not None and ord(char) in (0x002d, 0x002b):  # "-", "+"
                self._position += 1

            self._read_over_integer()

        end = self._position
        value = self._source[start:end]
        return (
            token.Float(start, end, value)
            if is_float
            else token.Integer(start, end, value)
        )

    def _read_over_integer(self):
        """ Advance lexer over an integer

        Returns:
            int: parsed value
        """
        char = self._peek()
        if char is None:
            raise UnexpectedEOF(self._position, self._source)

        code = ord(char)

        if code == 0x0030:  # "0"
            self._position += 1
            char = self._peek()
            if char is not None and ord(char) == 0x0030:
                raise UnexpectedCharacter(
                    'Unexpected character "%s"' % char,
                    self._position,
                    self._source,
                )
        else:
            self._read_over_digits()

    def _read_over_digits(self):
        """ Advance lexer over a sequence of digits """
        char = self._peek()
        if char is None:
            raise UnexpectedEOF(self._position, self._source)

        code = ord(char)
        if not (0x0030 <= code <= 0x0039):
            raise UnexpectedCharacter(
                'Unexpected character "%s"' % char, self._position, self._source
            )

        while True:
            char = self._peek()
            if char is not None and 0x0030 <= ord(char) <= 0x0039:
                self._position += 1
            else:
                break

    def _read_name(self):
        """ Advance lexer over a name ``/[_A-Za-z][A-Za-z0-9_]+/``.

        Returns:
            py_gql.lang.token.Name: parse token
        """
        start = self._position
        char = self._peek()
        while True:
            char = self._peek()
            if char is None:
                break

            code = ord(char)
            if (
                code == 0x005f
                or 0x0041 <= code <= 0x005a
                or 0x0061 <= code <= 0x007a
                or 0x0030 <= code <= 0x0039
            ):
                self._position += 1
            else:
                break

        end = self._position
        value = self._source[start:end]
        return token.Name(start, end, value)

    def __iter__(self):
        return self

    def __next__(self):
        """ Advance lexer and return the next :class:`py_gql.lang.token.Token`
        instance.

        Returns:
            py_gql.lang.token.Token: parse token

        Raises:
            :class:`~py_gql.exc.UnexpectedEOF`
            :class:`~py_gql.exc.InvalidCharacter`
            :class:`~py_gql.exc.UnexpectedCharacter`
            :class:`~py_gql.exc.NonTerminatedString`
        """
        if self._done:
            raise StopIteration()

        if not self._started:
            self._started = True
            return token.SOF(0, 0)

        self._read_over_whitespace()
        char = self._peek()

        if char is None:
            self._done = True
            return token.EOF(self._position, self._position)

        code = ord(char)

        if not (code >= 0x0020 or code == 0x0009):
            self._position += 1
            raise InvalidCharacter(char, self._position, self._source)

        if char in SYMBOLS:
            start = self._position
            self._position += 1
            return SYMBOLS[char](start, self._position)
        elif char == ".":
            return self._read_ellipsis()
        elif self._range(0, 3) == '"""':
            return self._read_block_string()
        elif char == '"':
            return self._read_string()
        elif code == 0x002d or 0x0030 <= code <= 0x0039:
            return self._read_number()
        elif (
            code == 0x005f
            or 0x0041 <= code <= 0x005a
            or 0x0061 <= code <= 0x007a
        ):
            return self._read_name()
        else:
            raise UnexpectedCharacter(
                'Unexpected character "%s"' % char, self._position, self._source
            )

    next = __next__  # Python 2 iterator interface



# -*- coding: utf-8 -*-
# pylint: disable = too-many-ancestors
""" All exceptions for this library are defined in the :mod:`py_gql.exc`
module. """

from ._string_utils import highlight_location, index_to_loc, stringify_path
from ._utils import cached_property


class GraphQLError(Exception):
    """ Base GraphQL exception from which all other inherit. """

    def __init__(self, message):
        super(GraphQLError, self).__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class GraphQLResponseError(GraphQLError):
    """ Implementors of this are suitable for usage in GraphQL responses and
    exposing to end users. """

    def to_dict(self):
        raise NotImplementedError()


class GraphQLSyntaxError(GraphQLResponseError):
    """ The GraphQL document is invalid.

    Args:
        message (str): Explanatory message
        position (int): 0-indexed position locating the syntax error
        source (str): Source string from which the syntax error originated

    Attributes:
        message (str): Explanatory message
        position (int): 0-indexed position locating the syntax error
        source (str): Source string from which the syntax error originated
    """

    def __init__(self, message, position, source):
        super(GraphQLSyntaxError, self).__init__(message)
        self.source = source
        self.position = position

    @cached_property
    def highlighted(self):
        """ str: Message followed by a view of the source document pointing at
        the exact location of the error. """
        if self.source is not None:
            return (
                self.message
                + " "
                + highlight_location(self.source, self.position)
            )
        return self.message

    def __str__(self):
        return self.highlighted

    def to_dict(self):
        """ Convert the exception to a dictionnary that can be serialized to
        JSON and exposed in a graphql response.

        Returns:
            dict: Dict representation of the error.
        """
        line, col = index_to_loc(self.source, self.position)
        return {
            "message": str(self),
            "locations": [{"line": line, "columne": col}],
        }


class InvalidCharacter(GraphQLSyntaxError):
    pass


class UnexpectedCharacter(GraphQLSyntaxError):
    pass


class UnexpectedEOF(GraphQLSyntaxError):
    """
    Args:
        position (int): 0-indexed position locating the syntax error
        source (str): Source string from which the syntax error originated
    """

    def __init__(self, position, source):
        super(UnexpectedEOF, self).__init__(
            "Unexpected <EOF>", position, source
        )


class NonTerminatedString(GraphQLSyntaxError):
    pass


class InvalidEscapeSequence(GraphQLSyntaxError):
    pass


class UnexpectedToken(GraphQLSyntaxError):
    pass


class GraphQLLocatedError(GraphQLResponseError):
    """ Error that can be traced back to specific position(s) in the source
    document.

    Args:
        message (str): Explanatory message
        nodes (Optional[List[py_gql.lang.ast.Node]]):
            Node or nodes relevant to the exception
        path (list): Location of the error during execution

    Attributes:
        message (str): Explanatory message
        nodes (List[py_gql.lang.ast.Node]): Node or nodes relevant to the exception
        path (list): Location of the error during execution
    """

    def __init__(self, message, nodes=None, path=None):
        super(GraphQLLocatedError, self).__init__(message)
        self.path = path
        self.nodes = nodes[:] if nodes else []

    def __str__(self):
        return self.message

    def to_dict(self):
        """ Convert the exception to a dictionnary that can be serialized to
        JSON and exposed in a graphql response.

        Returns:
            dict: Dict representation of the error.
        """
        kv = (
            ("message", str(self)),
            (
                "locations",
                [
                    {"line": line, "column": col}
                    for line, col in (
                        index_to_loc(node.source, node.loc[0])
                        for node in self.nodes
                        if node.loc and node.source
                    )
                ],
            ),
            ("path", self.path if self.path is not None else None),
        )
        return {k: v for k, v in kv if v}


class InvalidValue(GraphQLLocatedError, ValueError):
    pass


class UnknownEnumValue(InvalidValue):
    pass


class UnknownVariable(InvalidValue):
    pass


class ScalarSerializationError(GraphQLError, ValueError):
    pass


class ScalarParsingError(InvalidValue):
    pass


class SchemaError(GraphQLError):
    pass


class UnknownType(SchemaError, KeyError):
    pass


class ValidationError(GraphQLLocatedError):
    pass


class ExecutionError(GraphQLResponseError):
    """ Error that prevented execution.

    Args:
        message (str): Explanatory message

    Attributes:
        message (str): Explanatory message
    """

    def to_dict(self):
        """ Convert the exception to a dictionnary that can be serialized to
        JSON and exposed in a graphql response.

        Returns:
            dict: Dict representation of the error.
        """
        return {"message": str(self)}


class VariableCoercionError(GraphQLLocatedError):
    pass


class VariablesCoercionError(GraphQLError):
    """ Collection of multiple :class:`VariableCoercionError`.

    Args:
        errors (List[VariableCoercionError]): Wrapped errors

    Attributes:
        errors (List[VariableCoercionError]): Wrapped errors
    """

    def __init__(self, errors):
        super(VariablesCoercionError, self).__init__("%d errors" % len(errors))
        self.errors = errors

    def __str__(self):
        if len(self.errors) == 1:
            return str(self.errors[0])
        return ",\n".join([str(err) for err in self.errors])


class CoercionError(GraphQLLocatedError):
    def __init__(self, message, node=None, path=None, value_path=None):
        super(CoercionError, self).__init__(message, node, path)
        self.value_path = value_path

    def __str__(self):
        if self.value_path:
            return "%s at %s" % (self.message, stringify_path(self.value_path))
        return self.message


class MultiCoercionError(CoercionError):
    """ Collection of multiple :class:`CoercionError`.

    Args:
        errors (List[CoercionError]): Wrapped errors

    Attributes:
        errors (List[CoercionError]): Wrapped errors
    """

    def __init__(self, errors):
        super(MultiCoercionError, self).__init__("%d errors" % len(errors))
        self.errors = errors

    def __str__(self):
        if len(self.errors) == 1:
            return str(self.errors[0])
        return ",\n".join([str(err) for err in self.errors])


class ResolverError(GraphQLLocatedError):
    """ Raised when an expected error happens during field resolution.

    Subclass or raise this exception directly for use in your own
    resolvers in order for them to report errors instead of crashing execution.
    If your exception exposes an ``extensions`` attribute it will be included
    in the serialized version without the need to override :meth:`to_dict`.

    Args:
        message (str): Explanatory message
        nodes (Optional[Union[List[py_gql.lang.ast.Node], py_gql.lang.ast.Node]]):
            Node or nodes relevant to the exception
        path (list): Location of the error during execution
        extensions (Optional[dict]): Error extensions

    Attributes:
        message (str): Explanatory message
        nodes (List[py_gql.lang.ast.Node]): Node or nodes relevant to the exception
        path (list): Location of the error during execution
        extensions (Optional[dict]): Error extensions

    """

    def __init__(self, message, nodes=None, path=None, extensions=None):
        super(ResolverError, self).__init__(message, nodes, path)
        self.extensions = extensions

    def to_dict(self):
        """ Convert the exception to a dictionnary that can be serialized to
        JSON and exposed in a graphql response.

        Returns:
            dict: Dict representation of the error.
        """
        dict_ = super(ResolverError, self).to_dict()
        if self.extensions:
            dict_["extensions"] = dict(self.extensions)
        return dict_


class SDLError(GraphQLLocatedError):
    """ Error that occured when parsing and / or applying a schema definition
    document (SDL). """

    pass


class ExtensionError(SDLError):
    """ Error that occured when applying a schema or type extension node
    to an existing schema. """

    pass


class SchemaDirectiveError(SDLError):
    pass
